# -*- coding: iso-8859-1 -*-
"""
This module contains all the classes that allow to construct efficiently
low-thrust tajectories using our own flavour of the Sims-Flanagan model: a trajectory
transcription method that forms the basis for MALTO, the software in use in JPL
for preliminary interplanetary trajectory design.
"""
from pykep.util.util import *
from pykep import __extensions__


if __extensions__['numba']:
    from pykep.util.gravity_spherical_harmonic import gravity_spherical_harmonic
    from pykep.util.load_gravity_model import load_gravity_model


def read_satcat(satcatfilename=None):
    """
    pykep.read_satcat("my_dir/satcat.txt")

    This function reads the satcat catalogue, as can be downloaded from http://www.celestrak.com/NORAD/elements/
    and returns a dictionary keyed with the satellite international designator (e.g. "1958-002B"). Each entry
    is a named tuple with the following fields::

      noradn, multnameflag, payloadflag, operationstatus, name, ownership, launchdate, launchsite, decay, period, incl, apogee, perigee, radarA, orbitstatus

    see http://celestrak.com/satcat/satcat-format.asp for the meaning of all entries.

    Example::

      satcat = pykep.read_satcat("my_dir/satcat.txt")
      cross_section = satcat["1958-002B"].radarA
    """
    from collections import namedtuple
    satcatentry = namedtuple(
        'satcatentry', 'noradn multnameflag payloadflag operationstatus name ownership launchdate launchsite decay period incl apogee perigee radarA orbitstatus')
    satcat = dict()

    with open(satcatfilename, 'r') as f:
        for l1 in f:
            intdsgn = l1[0:11].strip()
            satcat[intdsgn] = satcatentry(l1[13:18], l1[19:20], l1[20:21], l1[21:22], l1[23:47], l1[49:54], l1[
                                          56:66], l1[68:73], l1[75:85], l1[87:94], l1[96:101], l1[103:109], l1[111:117], l1[119:127], l1[129:132])
    return satcat


def read_tle(tle_file, verbose=False, with_name=True):
    """
    planet_list = pykep.read_tle(tle_file, verbose=False, with_name=True)

    - tle_file: A string containin the file name (assumed to be in the working directory)
    - verbose: Activates some screen output to show the progress.
    - with_name: When True the TLE files contain three lines, the first one containing the satellite common name

    - [out] planet_list: a list of pykep tle_planets having as name their international designator (e.g. "1958-002B").

    This function reads a Two-Line-Element file as taken from the NORAD database
    http://www.celestrak.com/NORAD/elements/ or equivalent and returns a list of pykep planet_tle
    objects

    .. note::

       The name of each of the instantiated planets will be its international designator and
       thus a valid key to the satcat dictionary

    Example::

      planet_list = pykep.read_tle("my_dir/cosmos.tle")
      satcat = pykep.read_satcat("my_dir/satcat.txt")
      cross_sections = [satcat[pl.name()].radarA for pl in  planet_list]
    """
    from pykep.planet import tle
    planet_list = []

    with open(tle_file, 'r') as f:
        for line in f:
            if with_name:
                line1 = next(f)[:69]
            else:
                line1 = line[:69]
            line2 = next(f)[:69]
            planet_list.append(tle(line1, line2))
    return planet_list



"""
This module contains interplanetary trajectories problems in the form of pygmo.problem classes. The use
of pygmo.archipelago and pygmo.island with these may result in slow execution as the parallel execution will
be handled by python multiprocessing module, which is far slower than boost::threads (used instead when the pygmo problems
are implemented in c++)
"""
from pykep.trajopt._lt_margo import lt_margo
from pykep.trajopt._mga_1dsm import mga_1dsm
from pykep.trajopt._mga import mga
from pykep.trajopt._mga_lt_nep import mga_lt_nep
from pykep.trajopt._mr_lt_nep import mr_lt_nep
from pykep.trajopt._pl2pl_N_impulses import pl2pl_N_impulses
# The launchers models
from pykep.trajopt._launchers import _launchers
launchers = _launchers()
# The open traj gym
from . import gym 
from pykep.trajopt._indirect import indirect_or2or, indirect_pt2or, indirect_pt2pt, indirect_pt2pl
from pykep.trajopt._direct import direct_pl2pl



# -*- coding: iso-8859-1 -*-
"""
This module contains all the classes that allow to construct efficiently
low-thrust tajectories using our own flavour of the Sims-Flanagan model: a trajectory
transcription method that forms the basis for MALTO, the software in use in JPL
for preliminary interplanetary trajectory design.
"""
from pykep.planet.planet import *
from pykep.planet.planet import _base


def _keplerian_ctor(self, *args):
    """
pykep.planet.keplerian(when,orbital_elements, mu_central_body, mu_self,radius, safe_radius [, name = 'unknown'])

pykep.planet.keplerian(when,r,v, mu_central_body, mu_self,radius, safe_radius [, name = 'unknown'])

- when: a :py:class:`pykep.epoch` indicating the orbital elements reference epoch
- orbital_elements: a sequence of six containing a,e,i,W,w,M (SI units, i.e. meters and radiants)
- r,v: position and velocity of an object at when (SI units)
- mu_central_body: gravity parameter of the central body (SI units, i.e. m^2/s^3)
- mu_self: gravity parameter of the planet (SI units, i.e. m^2/s^3)
- radius: body radius (SI units, i.e. meters)
- safe_radius: mimimual radius that is safe during a fly-by of the planet (SI units, i.e. m)
- name: body name

.. note::

   All classes having Keplerian ephemerides as :py:class:`pykep.planet.mpcorb` inherit from this (c++) class

Example::

  earth = planet(epoch(54000,"mjd"),(9.99e-01 * AU, 1.67e-02, 8.85e-04 * DEG2RAD, 1.75e+02 * DEG2RAD, 2.87e+02 * DEG2RAD, 2.57e+02 * DEG2RAD), MU_SUN, 398600e9, 6378000, 6900000,  'Earth')"
    """
    self._orig_init(*args)
keplerian._orig_init = keplerian.__init__
keplerian.__init__ = _keplerian_ctor


def _j2_ctor(self, *args):
    """
pykep.planet.j2(when,orbital_elements, mu_central_body, mu_self,radius, safe_radius, J2RG2 [, name = 'unknown'])

pykep.planet.j2(when,r,v, mu_central_body, mu_self,radius, safe_radius, J2RG2 [, name = 'unknown'])

- when: a :py:class:`pykep.epoch` indicating the orbital elements reference epoch
- orbital_elements: a sequence of six containing a,e,i,W,w,M (SI units, i.e. meters and radiants)
- r,v: position and velocity of an object at when (SI units)
- mu_central_body: gravity parameter of the central body (SI units, i.e. m^2/s^3)
- mu_self: gravity parameter of the planet (SI units, i.e. m^2/s^3)
- radius: body radius (SI units, i.e. meters)
- safe_radius: mimimual radius that is safe during a fly-by of the planet (SI units, i.e. m)
- J2RG2: the product of J2 and the mean radius of the oblate primary squared (SI units, i.e. m^2)
- name: body name

Example::

    deb1 = planet.j2(epoch(54000,"mjd"),(7000000, 1.67e-02, 78.23 * DEG2RAD, 175. * DEG2RAD, 287. * DEG2RAD, 257 * DEG2RAD), MU_EARTH, 1, 1, 1, EARTH_J2*EARTH_RADIUS**2, 'deb1')
    """
    self._orig_init(*args)
j2._orig_init = j2.__init__
j2.__init__ = _j2_ctor



from pykep.sims_flanagan import spacecraft
from pykep.core import MU_SUN, EARTH_VELOCITY, G0, AU
import numpy as np


class _dynamics(object):

    def __init__(self, sc=spacecraft(1000, 0.3, 2500), mu=MU_SUN, alpha=1, bound=True):

        # check spacecraft
        if isinstance(sc, spacecraft):
            self.spacecraft = sc
        else:
            raise TypeError(
                "Spacecraft should be instance of pontryagin.spacecraft class.")

        # check gravitational parametre
        if not (isinstance(mu, float) or isinstance(mu, int)):
            raise TypeError(
                "Gravitational parametre, mu, must be supplied as either int or float.")
        elif not mu > 0:
            raise TypeError(
                "Gravitational parametre, mu, must be a positive number.")
        else:
            self.mu = float(mu)

        # check homotopy
        if not (isinstance(alpha, float) or isinstance(alpha, int)):
            raise TypeError(
                "Homotopy parametre, alpha, must be supplied as float or int.")
        elif not (alpha >= 0 and alpha <= 1):
            raise ValueError(
                "Homotopy parametre, alpha, must be between 0 and 1.")
        else:
            self.alpha = float(alpha)

        # check bound
        if not isinstance(bound, bool):
            raise TypeError(
                "Control bounding parametre, bound, supplied as boolean.")
        else:
            self.bound = bool(bound)

        # check control validity
        if (self.alpha == 1 and self.bound == False):
            raise ValueError(
                "Control can only be unbounded with quadratic control, i.e. bound == True if alpha == 1.")
        else:
            pass

        # spacecraft parametres
        self.c1 = self.spacecraft.thrust
        self.c2 = self.spacecraft.thrust / (self.spacecraft.isp * G0)

        # define nondimenional units
        self.L = AU
        self.V = EARTH_VELOCITY
        self.M = self.spacecraft.mass
        self.A = (self.V * self.V) / self.L
        self.F = self.M * self.A
        self.T = self.L / self.V
        self.Q = self.F / self.V

        # nondimensionalise parametres
        self.c1 /= self.F
        self.c2 /= self.Q
        self.mu /= MU_SUN

    def _eom_fullstate(self, fullstate):

        # extract state and control
        x, y, z, vx, vy, vz, m, lx, ly, lz, lvx, lvy, lvz, lvm, obj = fullstate
        u, ix, iy, iz = self._pontryagin(fullstate)

        # common subexpression elimination
        x0 = self.c1 * u / m
        x1 = x**2
        x2 = y**2
        x3 = z**2
        x4 = x1 + x2 + x3
        x5 = self.mu / x4**(3 / 2)
        x6 = x4**(-5 / 2)
        x7 = 3 * lvy * self.mu * x6 * y
        x8 = 3 * lvz * self.mu * x6 * z
        x9 = -x5
        x10 = 3 * self.mu * x6
        x11 = 3 * lvx * self.mu * x * x6
        x12 = self.c1 * u / m**2

        # fullstate transition
        dfs = np.array([
            vx,
            vy,
            vz,
            ix * x0 - x * x5,
            iy * x0 - x5 * y,
            iz * x0 - x5 * z,
            -self.c2 * u,
            -lvx * (x1 * x10 + x9) - x * x7 - x * x8,
            -lvy * (x10 * x2 + x9) - x11 * y - x8 * y,
            -lvz * (x10 * x3 + x9) - x11 * z - x7 * z,
            -lx,
            -ly,
            -lz,
            ix * lvx * x12 + iy * lvy * x12 + iz * lvz * x12,
            self.alpha * u + (1 - self.alpha) * u**2
        ])

        return dfs

    def _eom_fullstate_jac(self, fullstate):

        # extract state and control
        x, y, z, vx, vy, vz, m, lx, ly, lz, lvx, lvy, lvz, lvm = fullstate
        u, ix, iy, iz = self._pontryagin(fullstate)

        # common subexpression elimination
        x0 = x**2
        x1 = y**2
        x2 = z**2
        x3 = x0 + x1 + x2
        x4 = self.mu / x3**(3 / 2)
        x5 = -x4
        x6 = x3**(-5 / 2)
        x7 = 3 * self.mu * x6
        x8 = x0 * x7
        x9 = self.mu * x * x6
        x10 = 3 * x9
        x11 = x10 * y
        x12 = x10 * z
        x13 = self.c1 * u / m**2
        x14 = ix * x13
        x15 = x1 * x7
        x16 = self.mu * x6 * y
        x17 = 3 * x16
        x18 = x17 * z
        x19 = iy * x13
        x20 = x2 * x7
        x21 = iz * x13
        x22 = -lvy * x17
        x23 = self.mu * x6 * z
        x24 = 3 * x23
        x25 = -lvz * x24
        x26 = x3**(-7 / 2)
        x27 = 15 * self.mu * x0 * x26
        x28 = x27 * y
        x29 = x27 * z
        x30 = 15 * self.mu * x26
        x31 = 15 * self.mu * x * x26 * y * z
        x32 = lvz * x31
        x33 = 15 * self.mu * x * x26
        x34 = x1 * x33
        x35 = lvy * x31
        x36 = x2 * x33
        x37 = -x11
        x38 = -x12
        x39 = -lvx * x10
        x40 = x1 * x30 * z
        x41 = lvx * x31
        x42 = x2 * x30 * y
        x43 = -x18
        x44 = 2 * self.c1 * u / m**3

        # fullstate transition jacobian
        dfsj = np.array([
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            [x5 + x8, x11, x12, 0, 0, 0, -x14, 0, 0, 0, 0, 0, 0, 0],
            [x11, x15 + x5, x18, 0, 0, 0, -x19, 0, 0, 0, 0, 0, 0, 0],
            [x12, x18, x20 + x5, 0, 0, 0, -x21, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [-lvx * (-x**3 * x30 + 9 * x9) + lvy * x28 + lvz * x29 + x22 + x25, -lvx * (x17 - x28) - lvy * x10 + lvy *
             x34 + x32, -lvx * (x24 - x29) - lvz * x10 + lvz * x36 + x35, 0, 0, 0, 0, 0, 0, 0, x4 - x8, x37, x38, 0],
            [-lvx * x17 + lvx * x28 - lvy * (x10 - x34) + x32, lvx * x34 - lvy * (9 * x16 - x30 * y**3) + lvz * x40 + x25 + x39, -lvy * (
                x24 - x40) - lvz * x17 + lvz * x42 + x41, 0, 0, 0, 0, 0, 0, 0, x37, -x15 + x4, x43, 0],
            [-lvx * x24 + lvx * x29 - lvz * (x10 - x36) + x35, -lvy * x24 + lvy * x40 - lvz * (x17 - x42) + x41, lvx *
             x36 + lvy * x42 - lvz * (9 * x23 - x30 * z**3) + x22 + x39, 0, 0, 0, 0, 0, 0, 0, x38, x43, -x20 + x4, 0],
            [0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, -lvx * ix * x44 - lvy * iy *
                x44 - lvz * iz * x44, 0, 0, 0, x14, x19, x21, 0]
        ])

        return dfsj

    def _hamiltonian(self, fullstate):

        # extract fullstate and control
        x, y, z, vx, vy, vz, m, lx, ly, lz, lvx, lvy, lvz, lm, obj = fullstate
        u, ix, iy, iz = self._pontryagin(fullstate)

        # common subexpression elimination
        x0 = self.c1 * u / m
        x1 = self.mu / (x**2 + y**2 + z**2)**(3 / 2)

        # Hamiltonian
        H = -lm * self.c2 * u + lx * vx + ly * vy + lz * vz + lvx * (ix * x0 - x * x1) + lvy * (
            iy * x0 - x1 * y) + lvz * (iz * x0 - x1 * z) + self.alpha * u + u**2 * (-self.alpha + 1)

        return H

    def _pontryagin(self, fullstate):

        # extract fullstate
        x, y, z, vx, vy, vz, m, lx, ly, lz, lvx, lvy, lvz, lm, obj = fullstate

        # magnitude of lv
        lv = (lvx**2 + lvy**2 + lvz**2)**0.5

        # if mass-optimal control
        if self.alpha == 1:

            # switching function
            s = 1 - self.c1 * lv / m - self.c2 * lm

            # bang-bang control
            if s >= 0:
                u = 0.

            elif s < 0:
                u = 1.

        # if quadratic control
        else:
            u = (self.c1 * lv + m * (self.c2 * lm - self.alpha)) / \
                (2 * m * (1 - self.alpha))

            # if throttle is bounded
            if self.bound:
                u = max(u, 0.0)
                u = min(u, 1.0)

            elif not self.bound:
                pass

        # throttle direction
        ix = -lvx / lv
        iy = -lvy / lv
        iz = -lvz / lv

        # assemble control decision
        control = np.array([u, ix, iy, iz])

        return control



"""
This module contains several examples on the use of pykep, try one of the
run_example() methods!!!
"""

from pykep import __extensions__
from ._ex2 import run_example2

if __extensions__['pygmo']:
    from ._ex1 import run_example1
    from ._ex3 import run_example3
    from ._ex4 import run_example4
    from ._ex5 import run_example5
    from ._ex6 import run_example6
    from ._ex7 import run_example7
    from ._ex8 import run_example8
    from ._ex9 import run_example9
    from ._ex10 import run_example10
    from ._ex11 import run_example11
    from ._ex_utilities import add_gradient, algo_factory



import pykep as pk
import numpy as np

class _direct_base(object):
    """Base class for direct trajectory optimisation problems with one only leg.

    All inheriting classes will adopt, ``plot_traj``, ``plot_control``, and ``get_traj``.
    """

    def __init__(self, mass=1000., thrust=0.3, isp=3000., nseg=10, mu=pk.MU_SUN, hf=False):

        # segements
        if isinstance(nseg, int):
            self.nseg = nseg
        else:
            raise TypeError("nseg must be supplied as int.")

        # spacecraft
        self.sc = pk.sims_flanagan.spacecraft(mass, thrust, isp)

        # leg
        self.leg = pk.sims_flanagan.leg()
        self.leg.set_spacecraft(self.sc)
        self.leg.set_mu(mu)
        self.leg.high_fidelity = hf

    def get_nobj(self):
        return 1

    def get_nec(self):
        return 7

    def get_nic(self):
        return self.nseg

    def plot_traj(self, z, units=pk.AU, N=20, axes=None):
        """This function plots the 3 dimensional spacecraft trajectory, given a solution chromosome.

        Args:
            - z (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).
            - units (``float``, ``int``): units by which to scale the trajectory dimensions.
            - N (``int``): Number of points to be plotted along one arc.
        """
        import matplotlib.pyplot as plt
        # a call to the fitness on the chromosome z will change the class data member leg and set it
        # to represent the data in the chromosome z
        self.fitness(z)

        # creates a figure if needed
        if axes is None:
            fig = plt.figure()
            axes = fig.gca(projection='3d')

        # plots a small Sun
        axes.scatter([0], [0], [0], color='y')

        # plots the leg
        pk.orbit_plots.plot_sf_leg(self.leg, units=units, N=20, axes=axes)

        # plots problem specifics
        self._plot_traj(z, axes, units)

        return axes

    def plot_control(self, z, mark="k.-", time=True, axes=None):
        """Plots the control profile of the trajectory, as a function of time.

        Args:
            - z (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).
            - mark (``string``): matplotlib marker.
            - time (``bool``): If ``True``, x-axis is time in ``mjd2000``. If ``False``, x-axis is node index.

        """
        import matplotlib.pyplot as plt
        # data
        traj = self.get_traj(z)

        # time
        t = traj[:, 0] - traj[0, 0]

        # throttle
        u = traj[:, 8]

        # figure
        if axes is None:
            plt.figure()
            axes = plt.gca()

        # with time
        if time:
            axes.plot(t, u, mark)
            plt.xlabel("Time [days]")

        # without time
        elif not time:
            axes.plot(u, mark)
            plt.xlabel("Segment number")
        # no se
        else:
            raise RuntimeError("Something is wrong!")

        # label
        plt.ylabel("Throttle [ND]")

        return axes

    def get_traj(self, z):
        """Retrieves the trajectory information.
        The returned np.array contains::

            traj = [[t0, x0, y0, z0, vx0, vy0, vz0, m0, u0, ux0, uy0, uz0]
                    ...
                    [tf, xf, yf, zf, vxf, vyf, vzf, mf, uf, uxf, uyf, uzf]]

        Args:
            - z (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).

        Returns:
            np.array full information on states and controls along the trajectory nodes
        """

        # set leg
        self.fitness(z)

        # get states
        x = list(self.leg.get_states())
        # clean states
        for i in range(len(x)):
            # remove matchpoint duplicate
            x[i].pop(self.nseg)
            # convert to numpy.ndarray
            x[i] = np.asarray(x[i], np.float64)
            # time and mass
            if i in [0, 3]:
                x[i].reshape((self.nseg * 2 + 1, 1))
            # position and velocity
            elif i in [1, 2]:
                x[i].reshape((self.nseg * 2 + 1, 3))
            else:
                raise RuntimeError("Something is wrong!")

        # unpack states
        t, r, v, m = x

        # control
        u = self._get_controls(z)
        # since controls are only defined at midpoints we need to add the
        # values at the non midpoint nodes
        tmp = [0] * len(u) * 2
        for i in range(self.nseg):
            tmp[i * 6] = u[i * 3]
            tmp[i * 6 + 1] = u[i * 3 + 1]
            tmp[i * 6 + 2] = u[i * 3 + 2]
            tmp[i * 6 + 3] = u[i * 3]
            tmp[i * 6 + 4] = u[i * 3 + 1]
            tmp[i * 6 + 5] = u[i * 3 + 2]
        tmp.append(u[-3])
        tmp.append(u[-2])
        tmp.append(u[-1])
        u = np.asarray(tmp, np.float64).reshape((self.nseg * 2 + 1, 3))
        # throttle
        umag = np.linalg.norm(u, axis=1).reshape((self.nseg * 2 + 1, 1))

        # full dataset [t, x, y, z, vx, vy, vz, m, u, ux, uy, uz]
        return np.hstack((t.reshape((self.nseg * 2 + 1, 1)), r, v, m.reshape((self.nseg * 2 + 1, 1)), umag, u))

    def pretty(self, z):
        """
        pretty(x)

        Args:
            - x (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).

        Prints human readable information on the trajectory represented by the decision vector x
        """
        data = self.get_traj(z)
        self._pretty(z)

        print("\nSpacecraft Initial Position (m)  : [{!r}, {!r}, {!r}]".format(
            data[0, 1], data[0, 2], data[0, 3]))
        print("Spacecraft Initial Velocity (m/s): [{!r}, {!r}, {!r}]".format(
            data[0, 4], data[0, 5], data[0, 6]))
        print("Spacecraft Initial Mass  (kg)    : {!r}".format(data[0, 7]))

        print("Spacecraft Final Position (m)  : [{!r}, {!r}, {!r}]".format(
            data[-1, 1], data[-1, 2], data[-1, 3]))
        print("Spacecraft Final Velocity (m/s): [{!r}, {!r}, {!r}]".format(
            data[-1, 4], data[-1, 5], data[-1, 6]))
        print("Spacecraft Final Mass  (kg)    : {!r}".format(data[-1, 7]))


class direct_pl2pl(_direct_base):
    """Represents a direct transcription transfer between solar system planets.

    This problem works by manipulating the starting epoch t0, the transfer time T the final mass mf and the controls 
    The dicision vector is::

        z = [t0, T, mf, Vxi, Vyi, Vzi, Vxf, Vyf, Vzf, controls]
    """

    def __init__(self,
                 p0="earth",
                 pf="mars",
                 mass=1000,
                 thrust=0.3,
                 isp=3000,
                 nseg=20,
                 t0=[500, 1000],
                 tof=[200, 500],
                 vinf_dep=1e-3,
                 vinf_arr=1e-3,
                 hf=False):
        """Initialises a direct transcription orbit to orbit problem.

        Args:
            - p0 (``str``): Departure planet name. (will be used to construct a planet.jpl_lp object)
            - pf (``str``): Arrival planet name. (will be used to construct a planet.jpl_lp object)
            - mass (``float``, ``int``): Spacecraft wet mass [kg].
            - thrust (``float``, ``int``): Spacecraft maximum thrust [N].
            - isp (``float``, ``int``): Spacecraft specific impulse [s].
            - nseg (``int``): Number of colocation nodes.
            - t0 (``list``): Launch epochs bounds [mjd2000].
            - tof (``list``): Transfer time bounds [days].
            - vinf_dep (``float``): allowed launch DV [km/s] 
            - vinf_arr (``float``): allowed arrival DV [km/s]
            - hf (``bool``): High-fidelity. Activates a continuous representation for the thrust.
        """

        # initialise base
        _direct_base.__init__(self, mass, thrust, isp, nseg, pk.MU_SUN, hf)

        # planets
        if all([isinstance(pl, str) for pl in [p0, pf]]):
            self.p0 = pk.planet.jpl_lp(p0)
            self.pf = pk.planet.jpl_lp(pf)
        else:
            raise TypeError("Planet names must be supplied as str.")

        # bounds TODO check
        self.t0 = t0
        self.tof = tof

        # boundary conditions on velocity
        self.vinf_dep = vinf_dep * 1000  # (in m)
        self.vinf_arr = vinf_arr * 1000  # (in m)

        # The class is built around solar system plaents hence mu is always the
        # SUN
        self.mu = pk.MU_SUN

    def fitness(self, z):

        # epochs (mjd2000)
        t0 = pk.epoch(z[0])
        tf = pk.epoch(z[0] + z[1])

        # final mass
        mf = z[2]

        # controls
        u = z[9:]

        # compute Cartesian states of planets
        r0, v0 = self.p0.eph(t0)
        rf, vf = self.pf.eph(tf)

        # add the vinfs from the chromosome
        v0 = [a + b for a, b in zip(v0, z[3:6])]
        vf = [a + b for a, b in zip(vf, z[6:9])]

        # spacecraft states
        x0 = pk.sims_flanagan.sc_state(r0, v0, self.sc.mass)
        xf = pk.sims_flanagan.sc_state(rf, vf, mf)

        # set leg
        self.leg.set(t0, x0, u, tf, xf)

        # compute equality constraints
        ceq = np.asarray(self.leg.mismatch_constraints(), np.float64)

        # nondimensionalise equality constraints
        ceq[0:3] /= pk.AU
        ceq[3:6] /= pk.EARTH_VELOCITY
        ceq[6] /= self.sc.mass

        # compute inequality constraints
        cineq = np.asarray(self.leg.throttles_constraints(), np.float64)

        # compute inequality constraints on departure and arrival velocities
        v_dep_con = (z[3] ** 2 + z[4] ** 2 + z[5] ** 2 - self.vinf_dep ** 2)
        v_arr_con = (z[6] ** 2 + z[7] ** 2 + z[8] ** 2 - self.vinf_arr ** 2)

        # nondimensionalize inequality constraints
        v_dep_con /= pk.EARTH_VELOCITY ** 2
        v_arr_con /= pk.EARTH_VELOCITY ** 2

        return np.hstack(([-mf], ceq, cineq, [v_dep_con, v_arr_con]))

    def get_nic(self):
        return super().get_nic() + 2

    def get_bounds(self):
        lb = [self.t0[0], self.tof[0], self.sc.mass * 0.1] + \
            [-self.vinf_dep] * 3 + [-self.vinf_arr] * 3 + \
            [-1, -1, -1] * self.nseg
        ub = [self.t0[1], self.tof[1], self.sc.mass] + \
            [self.vinf_dep] * 3 + [self.vinf_arr] * 3 + \
            [1, 1, 1] * self.nseg
        return (lb, ub)

    def _plot_traj(self, z, axis, units):

        # times
        t0 = pk.epoch(z[0])
        tf = pk.epoch(z[0] + z[1])

        # plot Keplerian
        pk.orbit_plots.plot_planet(
            self.p0, t0, units=units, color=(0.8, 0.8, 0.8), axes=axis)
        pk.orbit_plots.plot_planet(
            self.pf, tf, units=units, color=(0.8, 0.8, 0.8), axes=axis)

    def _pretty(self, z):
        print("\nLow-thrust NEP transfer from " +
              self.p0.name + " to " + self.pf.name)
        print("\nLaunch epoch: {!r} MJD2000, a.k.a. {!r}".format(
            z[0], pk.epoch(z[0])))
        print("Arrival epoch: {!r} MJD2000, a.k.a. {!r}".format(
            z[0] + z[1], pk.epoch(z[0] + z[1])))
        print("Time of flight (days): {!r} ".format(z[1]))
        print("\nLaunch DV (km/s) {!r} - [{!r},{!r},{!r}]".format(np.sqrt(
            z[3]**2 + z[4]**2 + z[5]**2) / 1000, z[3] / 1000, z[4] / 1000, z[5] / 1000))
        print("Arrival DV (km/s) {!r} - [{!r},{!r},{!r}]".format(np.sqrt(
            z[6]**2 + z[7]**2 + z[8]**2) / 1000, z[6] / 1000, z[7] / 1000, z[8] / 1000))

    @staticmethod
    def _get_controls(z):
        return z[9:]


class direct_or2or(_direct_base):
    """Represents a direct transcription transfer between orbits.

    This problem works by manipulating the time of flight ``T``, final mass ``mf`` and mean anomolies ``M0, Mf``.
    ::

        z = [T, mf, M0, Mf, controls]
    """

    def __init__(self, elem0, elemf, mass, thrust, isp, nseg, Tlb, Tub, E0lb, E0ub, Eflb, Efub, mu=pk.MU_SUN, hf=True):
        """Initialises a direct transcription orbit to orbit problem.

        Args:
            - elem0 (``list``, ``tuple``, ``numpy.ndarray``): Departure Keplerian elements. The eccentric anomlay will be manipulated.
            - elemf (``list``, ``tuple``, ``numpy.ndarray``): Arrival Keplerian elements. The eccentric anomlay will be manipulated.
            - mass (``float``, ``int``): Spacecraft wet mass [kg].
            - thrust (``float``, ``int``): Spacecraft maximum thrust [N].
            - isp (``float``, ``int``): Spacecraft specific impulse [s].
            - nseg (``int``): Number of colocation nodes.
            - Tlb (``float``, ``int``): Minimimum time of flight [mjd2000].
            - Tub (``float``, ``int``): Maximum time of flight [mjd2000].
            - E0lb (``float``, ``int``): Minimum departure eccentric anomoly [rad].
            - E0ub (``float``, ``int``): Maximum departure eccentric anomoly [rad].
            - Eflb (``float``, ``int``): Minimum arrival eccentric anomoly [rad].
            - E0fb (``float``, ``int``): Maximum arrival eccentric anomoly [rad].
            - mu (``float``): Gravitational parameter of primary body [m^3/s^2].
            - hf (``bool``): ``True`` for continous thrust, ``False`` for impulsive thrust.
        """

        if all([(isinstance(elem, list) or isinstance(elem, tuple) or isinstance(elem, np.ndarray)) for elem in [elem0, elemf]]):
            elem0 = np.asarray(elem0, np.float64)
            elemf = np.asarray(elemf, np.float64)
        else:
            raise ValueError(
                "Both elem0 and elemf must be supplied as instances of list, tuple, or numpy.ndarray.")

        if all([elem.size == 6 for elem in [elem0, elemf]]):
            self.elem0 = elem0
            self.elemf = elemf
        else:
            raise TypeError("Both elem0 and elemf must be 6-dimensional.")

        if not all([(isinstance(T, float) or isinstance(T, int)) for T in [Tlb, Tub]]):
            raise TypeError(
                "Both Tlb and Tub must be supplied as instances of either float or int.")
        elif not Tlb < Tub:
            raise ValueError("Tlb must be less than Tub.")
        else:
            self.Tlb = float(Tlb)
            self.Tub = float(Tub)

        # initialise base
        _direct_base.__init__(self, mass, thrust, isp, nseg, mu, hf)

        self.mu = mu
        self.E0lb = E0lb
        self.E0ub = E0ub
        self.Eflb = Eflb
        self.Efub = Efub

    def fitness(self, z):

        # epochs (mjd2000)
        t0 = pk.epoch(0)
        tf = pk.epoch(z[0])

        # final mass
        mf = z[1]

        # eccentric anomolies
        E0 = z[2]
        Ef = z[3]

        # controls
        u = z[4:]

        # set Keplerian elements
        self.elem0[5] = E0
        self.elemf[5] = Ef

        # compute Cartesian states
        r0, v0 = pk.par2ic(self.elem0, self.mu)
        rf, vf = pk.par2ic(self.elemf, self.mu)

        # spacecraft states
        x0 = pk.sims_flanagan.sc_state(r0, v0, self.sc.mass)
        xf = pk.sims_flanagan.sc_state(rf, vf, mf)

        # set leg
        self.leg.set(t0, x0, u, tf, xf)

        # compute equality constraints
        ceq = np.asarray(self.leg.mismatch_constraints(), np.float64)

        # nondimensionalise equality constraints
        ceq[0:3] /= pk.AU
        ceq[3:6] /= pk.EARTH_VELOCITY
        ceq[6] /= self.sc.mass

        # compute inequality constraints
        cineq = np.asarray(self.leg.throttles_constraints(), np.float64)

        return np.hstack(([-mf], ceq, cineq))

    def get_bounds(self):
        pi = 3.14159265359
        lb = [self.Tlb, self.sc.mass / 10, self.E0lb,
              self.Eflb, *(-1, -1, -1) * self.nseg]
        ub = [self.Tub, self.sc.mass, self.E0ub,
              self.Efub, *(1, 1, 1) * self.nseg]
        return (lb, ub)

    def _plot_traj(self, z, axis, units):

        # times
        t0 = pk.epoch(0)
        tf = pk.epoch(z[0])

        # Keplerian
        kep0 = pk.planet.keplerian(t0, self.elem0)
        kepf = pk.planet.keplerian(tf, self.elemf)

        # plot Keplerian
        pk.orbit_plots.plot_planet(
            kep0, t0, units=units, color=(0.8, 0.8, 0.8), axes=axis)
        pk.orbit_plots.plot_planet(
            kepf, tf, units=units, color=(0.8, 0.8, 0.8), axes=axis)

    @staticmethod
    def _get_controls(z):
        return z[4:]



###########################################################
# We check what extensions are available
###########################################################
__extensions__ = {'matplotlib': False, 'mplot3d': False,
                  'pygmo': False, 'scikit-learn': False, 'scipy': False,
                  'numba': False}
# 1 - matplotlib
try:
    from matplotlib import __version__ as matplotlib_ver
    __extensions__['matplotlib'] = True

    # We detect the version and if more than 1.1.0 mplot3d is there
    mver = matplotlib_ver.split('.')
    mver = int(mver[0]) * 100 + int(mver[1]) * 10
    if mver >= 110:
        __extensions__['mplot3d'] = True
    del mver
except ImportError:
    pass

# 2 - pygmo2
try:
    from pygmo import __version__ as pygmo_ver
    __extensions__['pygmo'] = True
except ImportError:
    pass

# 3 - scikit-learn is installed
try:
    from sklearn import __version__ as sklearn_ver
    __extensions__['scikit-learn'] = True
except ImportError:
    pass

# 4 - scipy is installed
try:
    from scipy import __version__ as scipy_ver
    __extensions__['scipy'] = True
except ImportError:
    pass

# 5 - numba is installed
try:
    from numba import __version__ as numba_ver
    __extensions__['numba'] = True
except ImportError:
    pass

###########################################################
# We import the submodules
###########################################################

# For convenience, bring all core classes into the root namespace when
# importing *.
from pykep.core import *
from pykep import core, sims_flanagan, pontryagin, orbit_plots, examples, phasing, util, planet, trajopt


###########################################################
# We define pykep module
###########################################################
version = '2.4.1'
__doc__ = 'pykep is the answer ... but what was the question?'
__all__ = ['core', 'sims_flanagan', 'pontryagin', 'orbit_plots',
           'examples', 'trajopt', 'phasing', 'util', 'planet', 'test']
__version__ = {'major': int(version.split('.')[0]), 'minor': int(version.split('.')[1])}
__all__ += [name for name in dir(core) if not name.startswith('_')]



from pykep.core import epoch, lambert_problem, DAY2SEC, fb_vel, AU
from pykep.planet import jpl_lp

import numpy as np


class mga:
    r"""
    This class transcribes a Multiple Gravity Assist (MGA) trajectory with no deep space manouvres into an optimisation problem.
    It may be used as a User Defined Problem (UDP) for the pygmo (http://esa.github.io/pygmo/) optimisation suite.

    - Izzo, Dario. "Global optimization and space pruning for spacecraft trajectory design." Spacecraft Trajectory Optimization 1 (2010): 178-200.

    The decision vector (chromosome) is::

      direct encoding: [t0, T1, T2 ... ] in [mjd2000, days, days ... ]
      alpha encoding:  [t0, T, a1, a2 ...] in [mjd2000, days, nd, nd ... ]
      eta encoding:    [t0, n1, n2, n3 ...] in [mjd2000, nd, nd ...]

    .. note::

       The time of flights of a MGA trajectory (and in general) can be encoded in different ways.
       When they are directly present in the decision vector, we have the *direct* encoding. This is the most 'evolvable' encoding
       but also the one that requires the most problem knowledge (e.g. to define the bounds on each leg) and is not 
       very flexible in dealing with constraints on the total time of flight. The *alpha* and *eta* encodings, instead, allow
       to only specify bounds on the time of flight of the entire trajectory, and not on the single legs: a property that is attractive
       for multi-objective optimization, for example.

       In the *alpha* encoding each leg time-of-flight is decoded as follows, T_i = T log(alpha_i) / \sum_n(log(alpha_n)).
       In the *eta* encoding  each leg time-of-flight is decoded as follows, T_i = (tof_max - \sum_0^(i-1)(T_j)) * eta_i

       The chromosome dimension for the direct and eta encoding is the same, while the alpha encoding requires one more gene.

    .. note::

       The resulting problem is box-bounded (unconstrained).
    """

    def __init__(self,
                 seq=[jpl_lp('earth'), jpl_lp('venus'), jpl_lp('earth')],
                 t0=[0, 1000],
                 tof=[[30, 200], [200, 300]],
                 vinf=2.5,
                 multi_objective=False,
                 tof_encoding='direct',
                 orbit_insertion=False,
                 e_target=None,
                 rp_target=None
                 ):
        """mga(seq=[pk.planet.jpl_lp('earth'), pk.planet.jpl_lp('venus'), pk.planet.jpl_lp('earth')], t0=[0, 1000], tof=[100, 500], vinf=2.5, multi_objective=False, alpha_encoding=False, orbit_insertion=False, e_target=None, rp_target=None)

        Args:
            - seq (``list of pk.planet``): sequence of body encounters including the starting object
            - t0 (``list of pk.epoch``): the launch window
            - tof (``list`` or ``float``): bounds on the time of flight. If *tof_encoding* is 'direct', this contains a list
              of 2D lists defining the upper and lower bounds on each leg. If *tof_encoding* is 'alpha',
              this contains a 2D list with the lower and upper bounds on the time-of-flight. If *tof_encoding*
              is 'eta' tof is a float defining the upper bound on the time-of-flight
            - vinf (``float``): the vinf provided at launch for free
            - multi_objective (``bool``): when True constructs a multiobjective problem (dv, T). In this case, 'alpha' or `eta` encodings are reccomended
            - tof_encoding (``str``): one of 'direct', 'alpha' or 'eta'. Selects the encoding for the time of flights
            - orbit_insertion (``bool``): when True the arrival dv is computed as that required to acquire a target orbit defined by e_target and rp_target
            - e_target (``float``): if orbit_insertion is True this defines the target orbit eccentricity around the final planet
            - rp_target (``float``): if orbit_insertion is True this defines the target orbit pericenter around the final planet (in m)

        Raises:
            - ValueError: if *planets* do not share the same central body (checked on the mu_central_body attribute)
            - ValueError: if *t0* does not contain objects able to construct a epoch (e.g. pk. epoch or floats)
            - ValueError: if *tof* is badly defined
            - ValueError: it the target orbit is not defined and *orbit_insertion* is True
        """

        # Sanity checks
        # 1 - All planets need to have the same mu_central_body
        if ([r.mu_central_body for r in seq].count(seq[0].mu_central_body) != len(seq)):
            raise ValueError(
                'All planets in the sequence need to have exactly the same mu_central_body')
        # 2 - We try to build epochs out of the t0 list (mjd2000 by default)
        for i in range(len(t0)):
            if (type(t0[i]) != type(epoch(0))):
                t0[i] = epoch(t0[i])
        # 3 - Check the tof bounds
        if tof_encoding == 'alpha':
            if len(tof) != 2:
                raise ValueError(
                    r'When the tof_encoding is \'alpha\', tof is expected to be something like [lb, ub]')
        elif tof_encoding == 'direct':
            if len(tof) != (len(seq) - 1):
                raise ValueError(
                    'When tof_encoding is direct, the tof must be a float (upper bound on the time of flight)' + str(len(seq) - 1))
        elif tof_encoding == 'eta':
            try:
                float(tof)
            except TypeError:
                raise ValueError(
                    'The tof needs to be have len equal to  ' + str(len(seq) - 1))
        if not tof_encoding in ['alpha', 'eta', 'direct']:
            raise ValueError(
                "tof_encoding must be one of 'alpha', 'eta', 'direct'")

        # 4 - Check that if orbit insertion is selected e_target and r_p are
        # defined
        if orbit_insertion:
            if rp_target is None:
                raise ValueError(
                    'The rp_target needs to be specified when orbit insertion is selected')
            if e_target is None:
                raise ValueError(
                    'The e_target needs to be specified when orbit insertion is selected')

        # Public data members
        self.seq = seq
        self.t0 = t0
        self.tof = tof
        self.vinf = vinf * 1000
        self.multi_objective = multi_objective
        self.tof_encoding = tof_encoding
        self.orbit_insertion = orbit_insertion
        self.e_target = e_target
        self.rp_target = rp_target

        # Private data members
        self._n_legs = len(seq) - 1
        self._common_mu = seq[0].mu_central_body

    def get_nobj(self):
        return self.multi_objective + 1

    def get_bounds(self):
        t0 = self.t0
        tof = self.tof
        n_legs = self._n_legs

        if self.tof_encoding == 'alpha':
            # decision vector is  [t0, T, a1, a2, ....]
            lb = [t0[0].mjd2000, tof[0]] + [1e-3] * (n_legs)
            ub = [t0[1].mjd2000, tof[1]] + [1.0 - 1e-3] * (n_legs)
        elif self.tof_encoding == 'direct':
            # decision vector is  [t0, T1, T2, T3, ... ]
            lb = [t0[0].mjd2000] + [it[0] for it in self.tof]
            ub = [t0[1].mjd2000] + [it[1] for it in self.tof]
        elif self.tof_encoding == 'eta':
            # decision vector is  [t0, n1, n2, ....]
            lb = [t0[0].mjd2000] + [1e-3] * (n_legs)
            ub = [t0[1].mjd2000] + [1.0 - 1e-3] * (n_legs)
        return (lb, ub)

    def _decode_tofs(self, x):
        if self.tof_encoding == 'alpha':
            # decision vector is  [t0, T, a1, a2, ....]
            T = np.log(x[2:])
            return T / sum(T) * x[1]
        elif self.tof_encoding == 'direct':
            # decision vector is  [t0, T1, T2, T3, ... ]
            return x[1:]
        elif self.tof_encoding == 'eta':
            # decision vector is  [t0, n1, n2, n3, ... ]
            dt = self.tof
            T = [0] * self._n_legs
            T[0] = dt * x[1]
            for i in range(1, len(T)):
                T[i] = (dt - sum(T[:i])) * x[i + 1]
            return T

    def alpha2direct(self, x):
        """alpha2direct(x)

        Args:
            - x (``array-like``): a chromosome encoding an MGA trajectory in the alpha encoding

        Returns:
            ``numpy.array``: a chromosome encoding the MGA trajectory using the direct encoding
        """
        T = np.log(x[2:])
        retval = T / sum(T) * x[1]
        retval = np.insert(retval, 0, x[0])
        return retval

    def direct2alpha(self, x):
        """direct2alpha(x)

        Args:
            - x (``array-like``): a chromosome encoding an MGA trajectory in the direct encoding

        Returns:
            ``numpy.array``: a chromosome encoding the MGA trajectory using the alpha encoding
        """
        T = np.sum(x[1:])
        alphas = np.exp(x[1:] / (-T))
        retval = np.insert(alphas, 0, [x[0], T])
        return retval

    def eta2direct(self, x):
        """eta2direct(x)

        Args:
            - x (``array-like``): a chromosome encoding an MGA trajectory in the eta encoding

        Returns:
            ``numpy.array``: a chromosome encoding the MGA trajectory using the direct encoding

        Raises:
            - ValueError: when the tof_encoding is not 'eta'
        """
        if self.tof_encoding != 'eta':
            raise ValueError(
                "cannot call this method if the tof_encoding is not 'eta'")

        # decision vector is  [t0, n1, n2, n3, ... ]
        n = len(x) - 1
        dt = self.tof
        T = [0] * n
        T[0] = dt * x[1]
        for i in range(1, len(T)):
            T[i] = (dt - sum(T[:i])) * x[i + 1]
        np.insert(T, 0, [0])
        return T

    def direct2eta(self, x):
        """direct2eta(x)

        Args:
            - x (``array-like``): a chromosome encoding an MGA trajectory in the direct encoding

        Returns:
            ``numpy.array``: a chromosome encoding the MGA trajectory using the eta encoding

        Raises:
            - ValueError: when the tof_encoding is not 'eta'
        """
        if self.tof_encoding != 'eta':
            raise ValueError(
                "cannot call this method if the tof_encoding is not 'eta'")
        from copy import deepcopy
        retval = deepcopy(x)
        retval[1] = x[1] / self.tof
        for i in range(2, len(x)):
            retval[i] = x[i] / (self.tof - sum(x[1:i]))
        return retval

    def _compute_dvs(self, x):
        # 1 -  we 'decode' the times of flights and compute epochs (mjd2000)
        T = self._decode_tofs(x)  # [T1, T2 ...]
        ep = np.insert(T, 0, x[0])  # [t0, T1, T2 ...]
        ep = np.cumsum(ep)  # [t0, t1, t2, ...]
        # 2 - we compute the ephemerides
        r = [0] * len(self.seq)
        v = [0] * len(self.seq)
        for i in range(len(self.seq)):
            r[i], v[i] = self.seq[i].eph(ep[i])
        # 3 - we solve the lambert problems
        l = list()
        for i in range(self._n_legs):
            l.append(lambert_problem(
                r[i], r[i + 1], T[i] * DAY2SEC, self._common_mu, False, 0))
        # 4 - we compute the various dVs needed at fly-bys to match incoming
        # and outcoming
        DVfb = list()
        for i in range(len(l) - 1):
            vin = [a - b for a, b in zip(l[i].get_v2()[0], v[i + 1])]
            vout = [a - b for a, b in zip(l[i + 1].get_v1()[0], v[i + 1])]
            DVfb.append(fb_vel(vin, vout, self.seq[i + 1]))
        # 5 - we add the departure and arrival dVs
        DVlaunch_tot = np.linalg.norm(
            [a - b for a, b in zip(v[0], l[0].get_v1()[0])])
        DVlaunch = max(0, DVlaunch_tot - self.vinf)
        DVarrival = np.linalg.norm(
            [a - b for a, b in zip(v[-1], l[-1].get_v2()[0])])
        if self.orbit_insertion:
            # In this case we compute the insertion DV as a single pericenter
            # burn
            DVper = np.sqrt(DVarrival * DVarrival + 2 *
                            self.seq[-1].mu_self / self.rp_target)
            DVper2 = np.sqrt(2 * self.seq[-1].mu_self / self.rp_target -
                             self.seq[-1].mu_self / self.rp_target * (1. - self.e_target))
            DVarrival = np.abs(DVper - DVper2)
        return (DVlaunch, DVfb, DVarrival, l, DVlaunch_tot)

    # Objective function
    def fitness(self, x):
        DVlaunch, DVfb, DVarrival, _, _ = self._compute_dvs(x)
        if self.tof_encoding == 'direct':
            T = sum(x[1:])
        elif self.tof_encoding == 'alpha':
            T = x[1]
        elif self.tof_encoding == 'eta':
            T = sum(self.eta2direct(x)[1:])
        if self.multi_objective:
            return [DVlaunch + np.sum(DVfb) + DVarrival, T]
        else:
            return [DVlaunch + np.sum(DVfb) + DVarrival]

    def pretty(self, x):
        """pretty(x)

        Args:
            - x (``list``, ``tuple``, ``numpy.ndarray``): Decision chromosome, e.g. (``pygmo.population.champion_x``).

        Prints human readable information on the trajectory represented by the decision vector x
        """
        T = self._decode_tofs(x)
        ep = np.insert(T, 0, x[0])  # [t0, T1, T2 ...]
        ep = np.cumsum(ep)  # [t0, t1, t2, ...]
        DVlaunch, DVfb, DVarrival, l, DVlaunch_tot = self._compute_dvs(x)
        print("Multiple Gravity Assist (MGA) problem: ")
        print("Planet sequence: ", [pl.name for pl in self.seq])

        print("Departure: ", seq[0].name)
        print("\tEpoch: ", ep[0], " [mjd2000]")
        print("\tSpacecraft velocity: ", l[0].get_v1()[0], "[m/s]")
        print("\tHyperbolic velocity: ", DVlaunch_tot, "[m/s]")
        print("\tInitial DV: ", DVlaunch, "[m/s]")

        for pl, e, dv in zip(seq[1:-1], ep[1:-1], DVfb):
            print("Fly-by: ", pl.name)
            print("\tEpoch: ", e, " [mjd2000]")
            print("\tDV: ", dv, "[m/s]")

        print("Arrival: ", seq[-1].name)
        print("\tEpoch: ", ep[-1], " [mjd2000]")
        print("\tSpacecraft velocity: ", l[-1].get_v2()[0], "[m/s]")
        print("\tArrival DV: ", DVarrival, "[m/s]")

        print("Time of flights: ", T, "[days]")

    def plot(self, x, axes=None, units=AU, N=60):
        """plot(self, x, axes=None, units=pk.AU, N=60)

        Plots the spacecraft trajectory.

        Args:
            - x (``tuple``, ``list``, ``numpy.ndarray``): Decision chromosome.
            - axes (``matplotlib.axes._subplots.Axes3DSubplot``): 3D axes to use for the plot
            - units (``float``, ``int``): Length unit by which to normalise data.
            - N (``float``): Number of points to plot per leg
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        from pykep.orbit_plots import plot_planet, plot_lambert

        # Creating the axes if necessary
        if axes is None:
            mpl.rcParams['legend.fontsize'] = 10
            fig = plt.figure()
            axes = fig.gca(projection='3d')

        T = self._decode_tofs(x)
        ep = np.insert(T, 0, x[0])  # [t0, T1, T2 ...]
        ep = np.cumsum(ep)  # [t0, t1, t2, ...]
        _, _, _, l, _ = self._compute_dvs(x)
        for pl, e in zip(self.seq, ep):
            plot_planet(pl, epoch(e), units=units, legend=True,
                        color=(0.7, 0.7, 1), axes=axes)
        for lamb in l:
            plot_lambert(lamb, N=N, sol=0, units=units, color='k',
                         legend=False, axes=axes, alpha=0.8)
        return axes


if __name__ == "__main__":
    import pygmo as pg
    seq = [jpl_lp('earth'), jpl_lp('venus'), jpl_lp(
        'venus'), jpl_lp('earth'), jpl_lp('jupiter'), jpl_lp('saturn')]
    udp = mga(seq=seq,
              t0=[-1000., 0.],
              tof=[4000., 7000.],
              vinf=3.,
              tof_encoding='alpha',
              orbit_insertion=True,
              e_target=0.98,
              rp_target=108950000)

    udp = mga(seq=seq,
              t0=[-1000., 0.],
              tof=[[30, 400], [100, 470], [30, 400], [400, 2000], [1000, 6000]],
              vinf=3.,
              tof_encoding='direct',
              orbit_insertion=True,
              e_target=0.98,
              rp_target=108950000)

    udp = mga(seq=seq,
              t0=[-1000., 0.],
              tof=7000.,
              vinf=3.,
              tof_encoding='eta',
              orbit_insertion=True,
              e_target=0.98,
              rp_target=108950000)

    #udp = mga(seq=seq, t0=[-1000., 0.], tof=[[130,200], [430,470], [30, 70], [900, 1200], [4000, 5000]], vinf=0., alpha_encoding=False)
    prob = pg.problem(udp)
    uda = pg.cmaes(1500, force_bounds=True, sigma0=0.5, ftol=1e-4)
    #uda = pg.sade(4500)
    algo = pg.algorithm(uda)
    algo.set_verbosity(10)
    res = list()
    # for i in range(100):
    pop = pg.population(prob, 100)
    pop = algo.evolve(pop)
    res.append(pop.champion_f)



from ._tandem import tandem
from ._juice import juice, juice_mo
from ._cassini1 import cassini1, cassini1_a, cassini1_n
from ._rosetta import rosetta
from ._cassini2 import cassini2
from ._emNimp import em3imp, em5imp, em7imp
from ._eve_mga1dsm import eve_mga1dsm, eve_mga1dsm_a, eve_mga1dsm_n
from ._messenger import messenger




def run_example1(impulses=4):
    import pykep as pk
    import pygmo as pg
    import numpy as np
    from matplotlib import pyplot as plt
    from pykep.examples import add_gradient, algo_factory

    # problem
    udp = add_gradient(pk.trajopt.pl2pl_N_impulses(
        start=pk.planet.jpl_lp('earth'),
        target=pk.planet.jpl_lp('venus'),
        N_max=impulses,
        tof=[100., 1000.],
        vinf=[0., 4],
        phase_free=False,
        multi_objective=False,
        t0=[pk.epoch(0), pk.epoch(1000)]), with_grad=False)
    prob = pg.problem(udp)

    # algorithm
    uda = pg.cmaes(gen=1000, force_bounds = True)
    algo = pg.algorithm(uda)
    algo.set_verbosity(10)

    # population
    pop = pg.population(prob, 20)

    # solve the problem
    pop = algo.evolve(pop)

    # inspect the solution
    udp.udp_inner.plot(pop.champion_x)
    plt.ion()
    plt.show()

    udp.udp_inner.pretty(pop.champion_x)

if __name__ == "__main__":
    run_example1()



from numpy.linalg import norm
from math import sqrt, asin, acos

from pykep.core import epoch, fb_con, EARTH_VELOCITY, AU, MU_SUN
from pykep.planet import jpl_lp
from pykep.sims_flanagan import leg, spacecraft, sc_state

class mga_lt_nep:
    """
    This class is a pygmo (http://esa.github.io/pygmo/) problem representing a low-thrust
    interplanetary trajectory modelled as a Multiple Gravity Assist trajectory using multiple :class:`pykep.sims_flanagan_leg` 

    - Yam, C.H., di Lorenzo, D., and Izzo, D., Low-Thrust Trajectory Design as a Constrained Global Optimization Problem, Proceedings of the Institution of Mechanical Engineers, Part G: Journal of Aerospace Engineering, 225(11), pp.1243-1251, 2011.

    The decision vector (chromosome) is::

      [t0] + 
      [T1, mf1, Vxi1, Vyi1, Vzi1, Vxf1, Vyf1, Vzf1] + 
      [T2, mf2, Vxi2, Vyi2, Vzi2, Vxf2, Vyf2, Vzf2] + 
      ... + 
      [throttles1] + 
      [throttles2] + 
      ...

    .. note::

      The resulting optimization problem is non linearly constrained and unfeasible a.e. (almost everywhere)
    """

    def __init__(self,
                 seq = [jpl_lp('earth'), jpl_lp('venus'), jpl_lp('mercury')],
                 n_seg = [5, 20],
                 t0 = [3000, 4000],
                 tof = [[100, 1000], [200, 2000]],
                 vinf_dep = 3,
                 vinf_arr = 2,
                 mass = [1200., 2000.0],
                 Tmax = 0.5,
                 Isp = 3500.0,
                 fb_rel_vel = 6,
                 multi_objective = False,
                 high_fidelity = False):
        """
        Args:
            - seq (```list of pykep.planet```): defines the encounter sequence for the trajectoty (including the initial planet).
            - n_seg (```list``` of ```int```): the number of segments to be used for each leg.
            - t0 (```list``` of ```floats```): the launch window (in mjd2000).
            - tof (```list``` of ```2D-list```): minimum and maximum time of each leg (days).
            - vinf_dep (```float```): maximum launch hyperbolic velocity allowed (in km/sec).
            - vinf_arr (```float```): maximum arrival hyperbolic velocity allowed (in km/sec).
            - mass (```float```): spacecraft starting mass. (in kg).
            - Tmax (```float```): maximum thrust (in N).
            - Isp (```float```):: engine specific impulse (in s).
.           - fb_rel_vel (```float```): determines the bounds on the maximum allowed relative velocity at all fly-bys (in km/sec).
            - multi-objective (```bool```): when True defines the problem as a multi-objective problem, returning total DV and time of flight.
            - high_fidelity (```bool```):makes the trajectory computations slower, but actually dynamically feasible.
        """
        # We define some data members (we use the double underscore to
        # indicate they are private)
        self._seq = seq
        self._n_seg = n_seg
        self._t0 = t0
        self._tof = tof
        self._vinf_dep = vinf_dep * 1000
        self._vinf_arr = vinf_arr * 1000
        self._mass = mass
        self._Tmax = Tmax
        self._fb_rel_vel = fb_rel_vel
        self._multiobjective = multi_objective
        self._high_fidelity = high_fidelity

        self._n_legs = len(seq) - 1
        self._sc = spacecraft(mass[1], Tmax, Isp)
        self._leg = leg()
        self._leg.set_mu(MU_SUN)
        self._leg.set_spacecraft(self._sc)

    def get_bounds(self):
        # Convenience aliases
        t0 = self._t0
        mass = self._mass
        fb_rel_vel = self._fb_rel_vel
        n_legs = self._n_legs
        n_seg = self._n_seg
        vinf_dep = self._vinf_dep
        vinf_arr = self._vinf_arr
        tof = self._tof

        # Basic bounds
        lb = [t0[0]] + [0., mass[0], -fb_rel_vel, -fb_rel_vel, -fb_rel_vel, -
                                fb_rel_vel, -fb_rel_vel, -fb_rel_vel] * n_legs + [-1, -1, -1] * sum(n_seg)
        ub = [t0[1]] + [1, mass[1], fb_rel_vel, fb_rel_vel, fb_rel_vel, fb_rel_vel,
                                fb_rel_vel, fb_rel_vel] * n_legs + [1, 1, 1] * sum(n_seg)
        # bounds on the vinfs......
        lb[3:6] = [-vinf_dep] * 3
        ub[3:6] = [vinf_dep] * 3
        lb[-sum(n_seg) * 3 - 3:-sum(n_seg) * 3] = [-vinf_arr] * 3
        ub[-sum(n_seg) * 3 - 3:-sum(n_seg) * 3] = [vinf_arr] * 3
        # and for the time of flight
        lb[1:1 + 8 * n_legs:8] = [el[0] for el in tof]
        ub[1:1 + 8 * n_legs:8] = [el[1] for el in tof]

        return (lb, ub)

    def fitness(self, x):
        # The final mass is the fitness
        objfun = [-x[2 + 8 * (self._n_legs - 1)]]
        eq_c = []
        ineq_c = []
        if self._multiobjective:
            objfun = objfun + [sum(x[1:8*self._n_legs:8])]
        # We compute the epochs and ephemerides of the planetary encounters
        t_P = list([None] * (self._n_legs + 1))
        r_P = list([None] * (self._n_legs + 1))
        v_P = list([None] * (self._n_legs + 1))
        for i in range(len(self._seq)):
            t_P[i] = epoch(x[0] + sum(x[1:i*8:8]))
            r_P[i], v_P[i] = self._seq[i].eph(t_P[i])
        # We assemble the constraints. 
        # 1 - Mismatch Constraints
        for i in range(self._n_legs):
            # Departure velocity of the spacecraft in the heliocentric frame
            v0 = [a + b for a, b in zip(v_P[i], x[3 + 8 * i:6 + 8 * i])]
            if i==0:
                m0 = self._mass[1]
            else:
                m0 = x[2 + 8 * (i-1)]
            x0 = sc_state(r_P[i], v0, m0)
            vf = [a + b for a, b in zip(v_P[i+1], x[6 + 8 * i:9 + 8 * i])]
            xf = sc_state(r_P[i+1], vf, x[2 + 8 * i])
            idx_start = 1 + 8 * self._n_legs + sum(self._n_seg[:i]) * 3
            idx_end   = 1 + 8 * self._n_legs + sum(self._n_seg[:i+1]) * 3
            self._leg.set(t_P[i], x0, x[idx_start:idx_end], t_P[i+1], xf)
            mismatch = list(self._leg.mismatch_constraints())
            # Making the mismatch non dimensional (assumes an heliocentric interplanetary trajectory)
            mismatch[0] /= AU
            mismatch[1] /= AU
            mismatch[2] /= AU
            mismatch[3] /= EARTH_VELOCITY
            mismatch[4] /= EARTH_VELOCITY
            mismatch[5] /= EARTH_VELOCITY
            mismatch[6] /= self._mass[1]
            eq_c = eq_c + mismatch
            ineq_c = ineq_c + list(self._leg.throttles_constraints())

        # 2 - Fly-by constraints
        for i in range(self._n_legs - 1):
            DV_eq, alpha_ineq = fb_con(x[6 + i * 8:9 + i * 8], x[11 + i * 8:14 + i * 8], self._seq[i+1])
            eq_c = eq_c + [DV_eq / (EARTH_VELOCITY * EARTH_VELOCITY)]
            ineq_c = ineq_c + [alpha_ineq]

        # 3 - Departure and arrival Vinf
        # departure
        v_dep_con = (x[3] * x[3] + x[4] * x[4] + x[5] * x[5] -
                     self._vinf_dep * self._vinf_dep) / (EARTH_VELOCITY * EARTH_VELOCITY)
        # arrival
        n_fb = self._n_legs - 1
        v_arr_con = (x[6 + n_fb * 8] * x[6 + n_fb * 8] + x[7 + n_fb * 8] * x[7 + n_fb * 8] + x[8 + n_fb * 8] * x[8 + n_fb * 8] -
                     self._vinf_arr * self._vinf_arr) / (EARTH_VELOCITY * EARTH_VELOCITY)
        ineq_c = ineq_c + [v_dep_con, v_arr_con]

        return objfun + eq_c + ineq_c

    def get_nec(self):
        return self._n_legs * 7 + (self._n_legs - 1)

    def get_nic(self):
        return sum(self._n_seg) + (self._n_legs - 1) + 2


    # And this helps visualizing the trajectory
    def plot(self, x, axes=None):
        """
        ax = prob.plot(x, ax=None)

        - x: encoded trajectory
        - ax: matplotlib axis where to plot. If None figure and axis will be created
        - [out] ax: matplotlib axis where to plot

        Plots the trajectory represented by a decision vector x on the 3d axis ax

        Example::

          ax = prob.plot(x)
        """
        import matplotlib as mpl
        from mpl_toolkits.mplot3d import Axes3D
        import matplotlib.pyplot as plt
        from pykep.orbit_plots import plot_sf_leg, plot_planet

        # Creating the axis if necessary
        if axes is None:
            mpl.rcParams['legend.fontsize'] = 10
            fig = plt.figure()
            ax = fig.gca(projection='3d')
        else:
            ax = axes

        # Plotting the Sun ........
        ax.scatter([0], [0], [0], color=['y'])

         # We compute the epochs and ephemerides of the planetary encounters
        t_P = list([None] * (self._n_legs + 1))
        r_P = list([None] * (self._n_legs + 1))
        v_P = list([None] * (self._n_legs + 1))
        for i in range(len(self._seq)):
            t_P[i] = epoch(x[0] + sum(x[1:i*8:8]))
            r_P[i], v_P[i] = self._seq[i].eph(t_P[i])
            plot_planet(self._seq[i], t0 = t_P[i],
                        units=AU, legend=False, color=(0.7, 0.7, 0.7), s=30, axes=ax)

        # We assemble the constraints. 
        # 1 - Mismatch Constraints
        for i in range(self._n_legs):
            # Departure velocity of the spacecraft in the heliocentric frame
            v0 = [a + b for a, b in zip(v_P[i], x[3 + 8 * i:6 + 8 * i])]
            if i==0:
                m0 = self._mass[1]
            else:
                m0 = x[2 + 8 * (i-1)]
            x0 = sc_state(r_P[i], v0, m0)
            vf = [a + b for a, b in zip(v_P[i+1], x[6 + 8 * i:9 + 8 * i])]
            xf = sc_state(r_P[i+1], vf, x[2 + 8 * i])
            idx_start = 1 + 8 * self._n_legs + sum(self._n_seg[:i]) * 3
            idx_end   = 1 + 8 * self._n_legs + sum(self._n_seg[:i+1]) * 3
            self._leg.set(t_P[i], x0, x[idx_start:idx_end], t_P[i+1], xf)
            plot_sf_leg(self._leg, units=AU, N=10, axes=ax, legend=False)

        return axes

    def high_fidelity(self, status = False):
        """
        prob.high_fidelity(status = True)

        - status: either True or False (True sets high fidelity on)

        Sets the trajectory high fidelity mode

        Example::

          prob.high_fidelity(True)
        """
        # We set the propagation fidelity
        self._leg.high_fidelity = status


def run_example9():
    """
    This example demonstrates the indirect method (cartesian) on a point to planet variable time scenario.
    The starting conditions are taken from a run of the indirect method.
    """
    import pykep as pk
    import pygmo as pg
    import numpy as np
    from matplotlib import pyplot as plt
    from pykep.examples import add_gradient, algo_factory

    # 1 - Algorithm
    algo = algo_factory("snopt7")

    # 2 - Problem
    udp = add_gradient(pk.trajopt.indirect_pt2pl(
        x0=[44459220055.461708, -145448367557.6174, 1195278.0377499966,
            31208.214734303529, 9931.5012318647168, -437.07278242521573, 1000],
        t0=1285.6637861007277,
        pf="mars",
        thrust=0.1,
        isp=3000,
        mu=pk.MU_SUN,
        tof=[600, 720],
        alpha=0,    # quadratic control
        bound=True),
        with_grad=False
    )
    prob = pg.problem(udp)
    prob.c_tol = [1e-5] * prob.get_nc()

    # 3 - Population
    pop = pg.population(prob)
    z = np.hstack(([np.random.uniform(udp.udp_inner.tof[0],
                                      udp.udp_inner.tof[1])], 10 * np.random.randn(7)))
    pop.push_back(z)

    # 4 - Solve the problem (evolve)
    pop = algo.evolve(pop)

    homotopy_path = [0.2, 0.4, 0.6, 0.8, 0.9, 0.98, 0.99, 0.995, 1]
    for alpha in homotopy_path:
        z = pop.champion_x
        print("alpha is: ", alpha)
        udp = add_gradient(pk.trajopt.indirect_pt2pl(
            x0=[44459220055.461708, -145448367557.6174, 1195278.0377499966,
                31208.214734303529, 9931.5012318647168, -437.07278242521573, 1000],
            t0=1285.6637861007277,
            pf="mars",
            thrust=0.1,
            isp=3000,
            mu=pk.MU_SUN,
            tof=[600, 720],
            alpha=alpha,    # quadratic control
            bound=True),
            with_grad=False
        )
        prob = pg.problem(udp)
        prob.c_tol = [1e-5] * prob.get_nc()

        # 7 - Solve it
        pop = pg.population(prob)
        pop.push_back(z)
        pop = algo.evolve(pop)

    # 8 - Inspect the solution
    print("Feasible?:", prob.feasibility_x(pop.champion_x))

    # plot trajectory
    axis = udp.udp_inner.plot_traj(pop.champion_x, quiver=True, mark='k')
    plt.title("The trajectory in the heliocentric frame")

    # plot control
    udp.udp_inner.plot_control(pop.champion_x)
    plt.title("The control profile (throttle)")

    plt.ion()
    plt.show()

    udp.udp_inner.pretty(pop.champion_x)

    print("\nDecision vector: ", list(pop.champion_x))


if __name__ == "__main__":
    run_example9()



def plot_planet(plnt, t0='pykep.epoch(0)', N=60, units=1.0, color='k', alpha=1.0, s=40, legend=(False, False), axes=None):
    """
    ax = plot_planet(plnt, t0='pykep.epoch(0)', N=60, units=1.0, color='k', s=40, legend=False, axes=None)

    - axes:      3D axis object created using fig.gca(projection='3d')
    - plnt:      pykep.planet object we want to plot
    - t0:        a pykep.epoch or float (mjd2000) indicating when we want to plot the planet position
    - units:     the length unit to be used in the plot
    - color:     color to use to plot the orbit (passed to matplotlib)
    - s:         planet size (passed to matplotlib)
    - legend     tuple of two bool or string. The first element refers to the planet scatter plot, the second to the actual orbit.
                 If a bool value is used, then an automated legend label is generated (if True), if its a string the exact string is
                 used as a legend.

    Plots the planet position and its orbit.

    Example::

	import pykep as pk
	import matplotlib.pyplot as plt

	fig = plt.figure()
	ax = fig.gca(projection = '3d')
	pl = pk.planet.jpl_lp('earth')
	t_plot = pk.epoch(219)
	ax = pk.orbit_plots.plot_planet(pl, ax = ax, color='b')
    """
    from pykep import MU_SUN, SEC2DAY, epoch, AU, RAD2DEG
    from pykep.planet import keplerian
    from math import pi, sqrt
    import numpy as np
    import matplotlib.pylab as plt
    from mpl_toolkits.mplot3d import Axes3D

    if axes is None:
        fig = plt.figure()
        ax = fig.gca(projection='3d')
    else:
        ax = axes

    if t0 == 'pykep.epoch(0)':
        t0 = epoch(0)
    if type(t0) is not epoch:
        t0 = epoch(t0)

    # This is to make the tuple API compatible with the old API
    if legend is bool:
        legend = (legend, legend)

    # orbit period at epoch
    T = plnt.compute_period(t0) * SEC2DAY

    # make an osculating copy of the planet
    el = list(plnt.osculating_elements(t0))
    plnt_k = keplerian(t0, el, plnt.mu_central_body, plnt.mu_self, plnt.radius, plnt.safe_radius, plnt.name)

    # points where the orbit will be plotted
    when = np.linspace(0, T, N)

    # Ephemerides Calculation for the given planet
    x = np.array([0.0] * N)
    y = np.array([0.0] * N)
    z = np.array([0.0] * N)

    for i, day in enumerate(when):
        r, v = plnt_k.eph(epoch(t0.mjd2000 + day))
        x[i] = r[0] / units
        y[i] = r[1] / units
        z[i] = r[2] / units

    # Actual plot commands
    if (legend[0] is True):
        label1 = plnt.name + " " + t0.__repr__()[0:11]
    elif (legend[0] is False):
        label1 = None
    elif (legend[0] is None):
        label1 = None
    else:
        label1 = legend[0]

    if (legend[1] is True):
        label2 = plnt.name + " orbit"
    elif (legend[1] is False):
        label2 = None
    elif (legend[1] is None):
        label2 = None
    else:
        label2 = legend[1]

    ax.plot(x, y, z, label=label2, c=color, alpha=alpha)
    ax.scatter([x[0]], [y[0]], [z[0]], s=s, marker='o', alpha=0.8, c=[color], label=label1)

    if legend[0] or legend[1]:
        ax.legend()
    return ax


def plot_lambert(l, N=60, sol=0, units=1.0, color='b', legend=False, axes=None, alpha=1.):
    """
    ax = plot_lambert(l, N=60, sol=0, units='pykep.AU', legend='False', axes=None, alpha=1.)

    - axes:       3D axis object created using fig.gca(projection='3d')
    - l:        pykep.lambert_problem object
    - N:		number of points to be plotted along one arc
    - sol:      solution to the Lambert's problem we want to plot (must be in 0..Nmax*2)
                where Nmax is the maximum number of revolutions for which there exist a solution.
    - units:    the length unit to be used in the plot
    - color:    matplotlib color to use to plot the line
    - legend:   when True it plots also the legend with info on the Lambert's solution chosen

    Plots a particular solution to a Lambert's problem

    Example::

      import pykep as pk
      import matplotlib.pyplot as plt

      fig = plt.figure()
      ax = fig.gca(projection='3d')

      t1 = pk.epoch(0)
      t2 = pk.epoch(640)
      dt = (t2.mjd2000 - t1.mjd2000) * pk.DAY2SEC

      pl = pk.planet.jpl_lp('earth')
      pk.orbit_plots.plot_planet(pl, t0=t1, axes=ax, color='k')
      rE,vE = pl.eph(t1)

      pl = pk.planet.jpl_lp('mars')
      pk.orbit_plots.plot_planet(pl, t0=t2, axes=ax, color='r')
      rM, vM = pl.eph(t2)

      l = lambert_problem(rE,rM,dt,pk.MU_SUN)
      pk.orbit_plots.plot_lambert(l, ax=ax, color='b')
      pk.orbit_plots.plot_lambert(l, sol=1, axes=ax, color='g')
      pk.orbit_plots.plot_lambert(l, sol=2, axes=ax, color='g', legend = True)

      plt.show()
    """
    from pykep import propagate_lagrangian, AU
    import numpy as np
    import matplotlib.pylab as plt
    from mpl_toolkits.mplot3d import Axes3D

    if axes is None:
        fig = plt.figure()
        ax = fig.gca(projection='3d')
    else:
        ax = axes

    if sol > l.get_Nmax() * 2:
        raise ValueError("sol must be in 0 .. NMax*2 \n * Nmax is the maximum number of revolutions for which there exist a solution to the Lambert's problem \n * You can compute Nmax calling the get_Nmax() method of the lambert_problem object")

    # We extract the relevant information from the Lambert's problem
    r = l.get_r1()
    v = l.get_v1()[sol]
    T = l.get_tof()
    mu = l.get_mu()

    # We define the integration time ...
    dt = T / (N - 1)

    # ... and alocate the cartesian components for r
    x = np.array([0.0] * N)
    y = np.array([0.0] * N)
    z = np.array([0.0] * N)

    # We calculate the spacecraft position at each dt
    for i in range(N):
        x[i] = r[0] / units
        y[i] = r[1] / units
        z[i] = r[2] / units
        r, v = propagate_lagrangian(r, v, dt, mu)

    # And we plot
    if legend:
        label = 'Lambert solution (' + str((sol + 1) // 2) + ' revs.)'
    else:
        label = None
    ax.plot(x, y, z, c=color, label=label, alpha=alpha)

    if legend:
        ax.legend()

    return ax


def plot_kepler(r0, v0, tof, mu, N=60, units=1, color='b', label=None, axes=None):
    """
    ax = plot_kepler(r0, v0, tof, mu, N=60, units=1, color='b', legend=False, axes=None):

    - axes:     3D axis object created using fig.gca(projection='3d')
    - r0:       initial position (cartesian coordinates)
    - v0:		initial velocity (cartesian coordinates)
    - tof:		propagation time
    - mu:		gravitational parameter
    - N:		number of points to be plotted along one arc
    - units:	the length unit to be used in the plot
    - color:	matplotlib color to use to plot the line
    - label 	adds a label to the plotted arc.

    Plots the result of a keplerian propagation

    Example::

        import pykep as pk
        pi = 3.14
        pk.orbit_plots.plot_kepler(r0 = [1,0,0], v0 = [0,1,0], tof = pi/3, mu = 1)
    """

    from pykep import propagate_lagrangian
    import matplotlib.pylab as plt
    from mpl_toolkits.mplot3d import Axes3D

    if axes is None:
        fig = plt.figure()
        ax = fig.gca(projection='3d')
    else:
        ax = axes

    # We define the integration time ...
    dt = tof / (N - 1)

    # ... and calculate the cartesian components for r
    x = [0.0] * N
    y = [0.0] * N
    z = [0.0] * N

    # We calculate the spacecraft position at each dt
    for i in range(N):
        x[i] = r[0] / units
        y[i] = r[1] / units
        z[i] = r[2] / units
        r, v = propagate_lagrangian(r0, v0, dt, mu)

    # And we plot
    ax.plot(x, y, z, c=color, label=label)
    return ax


def plot_taylor(r0, v0, m0, thrust, tof, mu, veff, N=60, units=1, color='b', legend=False, axes=None):
    """
    ax = plot_taylor(r0, v0, m0, thrust, tof, mu, veff, N=60, units=1, color='b', legend=False, axes=None):

    - axes:		3D axis object created using fig.gca(projection='3d')
    - r0:		initial position (cartesian coordinates)
    - v0:		initial velocity (cartesian coordinates)
    - m0: 		initial mass
    - u:		cartesian components for the constant thrust
    - tof:		propagation time
    - mu:		gravitational parameter
    - veff:	the product Isp * g0
    - N:		number of points to be plotted along one arc
    - units:	the length unit to be used in the plot
    - color:	matplotlib color to use to plot the line
    - legend:	when True it plots also the legend

    Plots the result of a taylor propagation of constant thrust

    Example::

	import pykep as pk
	import matplotlib.pyplot as plt
	pi = 3.14

	fig = plt.figure()
	ax = fig.gca(projection = '3d')
	pk.orbit_plots.plot_taylor([1,0,0],[0,1,0],100,[1,1,0],40, 1, 1, N = 1000, axes = ax)
	plt.show()
    """

    from pykep import propagate_taylor
    import matplotlib.pyplot as plt

    if axes is None:
        fig = plt.figure()
        ax = fig.gca(projection='3d')
    else:
        ax = axes

    # We define the integration time ...
    dt = tof / (N - 1)

    # ... and calcuate the cartesian components for r
    x = [0.0] * N
    y = [0.0] * N
    z = [0.0] * N

    # We calculate the spacecraft position at each dt
    for i in range(N):
        x[i] = r[0] / units
        y[i] = r[1] / units
        z[i] = r[2] / units
        r, v, m = propagate_taylor(r0, v0, m0, u, dt, mu, veff, -10, -10)

    # And we plot
    if legend:
        label = 'constant thrust arc'
    else:
        label = None
    ax.plot(x, y, z, c=color, label=label)

    if legend:
        ax.legend()

    if axes is None:  # show only if axis is not set
        plt.show()
    return ax


def plot_taylor_disturbance(r0, v0, m0, thrust, disturbance, tof, mu, veff, N=60, units=1, color='b', legend=False, axes=None):
    """
    ax = plot_taylor_disturbance(r, v, m, thrust, disturbance, t, mu, veff, N=60, units=1, color='b', legend=False, axes=None):

    - axes:			3D axis object created using fig.gca(projection='3d')
    - r0:			initial position (cartesian coordinates)
    - v0:			initial velocity (cartesian coordinates)
    - m0: 			initial mass
    - thrust:		cartesian components for the constant thrust
    - disturbance:	cartesian components for a constant disturbance (will not affect mass)
    - tof:			propagation time
    - mu:			gravitational parameter
    - veff:			the product Isp * g0
    - N:			number of points to be plotted along one arc
    - units:		the length unit to be used in the plot
    - color:		matplotlib color to use to plot the line
    - legend:		when True it plots also the legend

    Plots the result of a taylor propagation of constant thrust
    """

    from pykep import propagate_taylor_disturbance
    import matplotlib.pyplot as plt

    if axes is None:
        fig = plt.figure()
        ax = fig.gca(projection='3d')
    else:
        ax = axes

    # We define the integration time ...
    dt = tof / (N - 1)

    # ... and calcuate the cartesian components for r
    x = [0.0] * N
    y = [0.0] * N
    z = [0.0] * N

    # We calculate the spacecraft position at each dt
    for i in range(N):
        x[i] = r[0] / units
        y[i] = r[1] / units
        z[i] = r[2] / units
        r0, v0, m0 = propagate_taylor_disturbance(
            r0, v0, m0, thrust, disturbance, dt, mu, veff, -10, -10)

    # And we plot
    if legend:
        label = 'constant thrust arc'
    else:
        label = None
    ax.plot(x, y, z, c=color, label=label)
    return ax


def plot_sf_leg(leg, N=5, units=1, color='b', legend=False, plot_line=True, plot_segments=False, axes=None):
    """
    ax = plot_sf_leg(leg, N=5, units=1, color='b', legend=False, no_trajectory=False, axes=None):

    - axes:		    3D axis object created using fig.gca(projection='3d')
    - leg:	        a pykep.sims_flanagan.leg
    - N:		    number of points to be plotted along one arc
    - units:	    the length unit to be used in the plot
    - color:	    matplotlib color to use to plot the trajectory and the grid points
    - legend	    when True it plots also the legend
    - plot_line: 	when True plots also the trajectory (between mid-points and grid points)

    Plots a Sims-Flanagan leg

    Example::

        from mpl_toolkits.mplot3d import Axes3D
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.gca(projection='3d')
        t1 = epoch(0)
        pl = planet_ss('earth')
        rE,vE = pl.eph(t1)
        plot_planet(pl,t0=t1, units=AU, axes=ax)

        t2 = epoch(440)
        pl = planet_ss('mars')
        rM, vM = pl.eph(t2)
        plot_planet(pl,t0=t2, units=AU, axes=ax)

        sc = sims_flanagan.spacecraft(4500,0.5,2500)
        x0 = sims_flanagan.sc_state(rE,vE,sc.mass)
        xe = sims_flanagan.sc_state(rM,vM,sc.mass)
        l = sims_flanagan.leg(t1,x0,[1,0,0]*5,t2,xe,sc,MU_SUN)

        plot_sf_leg(l, units=AU, axes=ax)
    """
    from pykep import propagate_lagrangian, AU, DAY2SEC, G0, propagate_taylor
    import numpy as np
    from scipy.linalg import norm
    from math import exp
    import matplotlib.pylab as plt
    from mpl_toolkits.mplot3d import Axes3D

    if axes is None:
        fig = plt.figure()
        ax = fig.gca(projection='3d')
    else:
        ax = axes

    # We compute the number of segments for forward and backward propagation
    n_seg = len(leg.get_throttles())
    fwd_seg = (n_seg + 1) // 2
    back_seg = n_seg // 2

    # We extract information on the spacecraft
    sc = leg.get_spacecraft()
    isp = sc.isp
    max_thrust = sc.thrust

    # And on the leg
    throttles = leg.get_throttles()
    mu = leg.get_mu()

    # Forward propagation

    # x,y,z contain the cartesian components of all points (grid+midpints)
    x = [0.0] * (fwd_seg * 2 + 1)
    y = [0.0] * (fwd_seg * 2 + 1)
    z = [0.0] * (fwd_seg * 2 + 1)

    state = leg.get_xi()

    # Initial conditions
    r = state.r
    v = state.v
    m = state.m
    x[0] = r[0] / units
    y[0] = r[1] / units
    z[0] = r[2] / units

    # We compute all points by propagation
    for i, t in enumerate(throttles[:fwd_seg]):
        dt = (t.end.mjd - t.start.mjd) * DAY2SEC
        alpha = min(norm(t.value), 1.0)
        # Keplerian propagation and dV application
        if leg.high_fidelity is False:
            dV = [max_thrust / m * dt * dumb for dumb in t.value]
            if plot_line:
                plot_kepler(r, v, dt / 2, mu, N=N, units=units,
                            color=(alpha, 0, 1 - alpha), axes=ax)
            r, v = propagate_lagrangian(r, v, dt / 2, mu)
            x[2 * i + 1] = r[0] / units
            y[2 * i + 1] = r[1] / units
            z[2 * i + 1] = r[2] / units
            # v= v+dV
            v = [a + b for a, b in zip(v, dV)]
            if plot_line:
                plot_kepler(r, v, dt / 2, mu, N=N, units=units,
                            color=(alpha, 0, 1 - alpha), axes=ax)
            r, v = propagate_lagrangian(r, v, dt / 2, mu)
            x[2 * i + 2] = r[0] / units
            y[2 * i + 2] = r[1] / units
            z[2 * i + 2] = r[2] / units
            m *= exp(-norm(dV) / isp / G0)
        # Taylor propagation of constant thrust u
        else:
            u = [max_thrust * dumb for dumb in t.value]
            if plot_line:
                plot_taylor(r, v, m, u, dt / 2, mu, isp * G0, N=N,
                            units=units, color=(alpha, 0, 1 - alpha), axes=ax)
            r, v, m = propagate_taylor(
                r, v, m, u, dt / 2, mu, isp * G0, -12, -12)
            x[2 * i + 1] = r[0] / units
            y[2 * i + 1] = r[1] / units
            z[2 * i + 1] = r[2] / units
            if plot_line:
                plot_taylor(r, v, m, u, dt / 2, mu, isp * G0, N=N,
                            units=units, color=(alpha, 0, 1 - alpha), axes=ax)
            r, v, m = propagate_taylor(
                r, v, m, u, dt / 2, mu, isp * G0, -12, -12)
            x[2 * i + 2] = r[0] / units
            y[2 * i + 2] = r[1] / units
            z[2 * i + 2] = r[2] / units

    x_grid = x[::2]
    y_grid = y[::2]
    z_grid = z[::2]
    x_midpoint = x[1::2]
    y_midpoint = y[1::2]
    z_midpoint = z[1::2]
    if plot_segments:
        ax.scatter(x_grid[:-1], y_grid[:-1], z_grid[:-1],
                     label='nodes', marker='o')
        ax.scatter(x_midpoint, y_midpoint, z_midpoint,
                     label='mid-points', marker='x')
        ax.scatter(x_grid[-1], y_grid[-1], z_grid[-1],
                     marker='^', c='y', label='mismatch point')

    # Backward propagation

    # x,y,z will contain the cartesian components of
    x = [0.0] * (back_seg * 2 + 1)
    y = [0.0] * (back_seg * 2 + 1)
    z = [0.0] * (back_seg * 2 + 1)

    state = leg.get_xf()

    # Final conditions
    r = state.r
    v = state.v
    m = state.m
    x[-1] = r[0] / units
    y[-1] = r[1] / units
    z[-1] = r[2] / units

    for i, t in enumerate(throttles[-1:-back_seg - 1:-1]):
        dt = (t.end.mjd - t.start.mjd) * DAY2SEC
        alpha = min(norm(t.value), 1.0)
        if leg.high_fidelity is False:
            dV = [max_thrust / m * dt * dumb for dumb in t.value]
            if plot_line:
                plot_kepler(r, v, -dt / 2, mu, N=N, units=units,
                            color=(alpha, 0, 1 - alpha), axes=ax)
            r, v = propagate_lagrangian(r, v, -dt / 2, mu)
            x[-2 * i - 2] = r[0] / units
            y[-2 * i - 2] = r[1] / units
            z[-2 * i - 2] = r[2] / units
            # v= v+dV
            v = [a - b for a, b in zip(v, dV)]
            if plot_line:
                plot_kepler(r, v, -dt / 2, mu, N=N, units=units,
                            color=(alpha, 0, 1 - alpha), axes=ax)
            r, v = propagate_lagrangian(r, v, -dt / 2, mu)
            x[-2 * i - 3] = r[0] / units
            y[-2 * i - 3] = r[1] / units
            z[-2 * i - 3] = r[2] / units
            m *= exp(norm(dV) / isp / G0)
        else:
            u = [max_thrust * dumb for dumb in t.value]
            if plot_line:
                plot_taylor(r, v, m, u, -dt / 2, mu, isp * G0, N=N,
                            units=units, color=(alpha, 0, 1 - alpha), axes=ax)
            r, v, m = propagate_taylor(
                r, v, m, u, -dt / 2, mu, isp * G0, -12, -12)
            x[-2 * i - 2] = r[0] / units
            y[-2 * i - 2] = r[1] / units
            z[-2 * i - 2] = r[2] / units
            if plot_line:
                plot_taylor(r, v, m, u, -dt / 2, mu, isp * G0, N=N,
                            units=units, color=(alpha, 0, 1 - alpha), axes=ax)
            r, v, m = propagate_taylor(
                r, v, m, u, -dt / 2, mu, isp * G0, -12, -12)
            x[-2 * i - 3] = r[0] / units
            y[-2 * i - 3] = r[1] / units
            z[-2 * i - 3] = r[2] / units

    x_grid = x[::2]
    y_grid = y[::2]
    z_grid = z[::2]
    x_midpoint = x[1::2]
    y_midpoint = y[1::2]
    z_midpoint = z[1::2]

    if plot_segments:
        ax.scatter(x_grid[1:], y_grid[1:], z_grid[
                     1:], marker='o', label='nodes')
        ax.scatter(x_midpoint, y_midpoint, z_midpoint,
                     marker='x', label='mid-points')
        ax.scatter(x_grid[0], y_grid[0], z_grid[0],
                     marker='^', c='y', label='mismatch point')

    if legend:
        ax.legend()

    if axes is None:  # show only if axis is not set
        plt.show()
    return ax



# We define a meta-problem augmenting a generic UDP with a numercial gradient so that algorithms
# needing the gradient can be used. Note that algorithm such as snopt7 have an internal system to compute gradients
# in case the problem is fed to it without. Hence the kwarg with_grad is
# provided to deactivate the gradient.

import pygmo as pg


class add_gradient:

    def __init__(self, udp, with_grad=False):
        self.udp_inner = udp
        self.prob = pg.problem(udp)
        self.with_grad = with_grad

    def fitness(self, x):
        return self.prob.fitness(x)

    def get_bounds(self):
        return self.prob.get_bounds()

    def get_nec(self):
        return self.prob.get_nec()

    def get_nic(self):
        return self.prob.get_nic()

    def get_nobj(self):
        return self.prob.get_nobj()

    def gradient(self, x):
        # we here use the low precision gradient
        return pg.estimate_gradient(lambda x: self.fitness(x), x, 1e-8)

    def has_gradient(self):
        return self.with_grad

# We define a utility class that sets the algorithm to be used (in pagmo's
# terminology the UDA or user defined algorithm).


def algo_factory(name, original_screen_output=True):
    if name is "slsqp":
        uda = pg.nlopt('slsqp')
        uda.xtol_rel = 1e-5
        uda.ftol_rel = 0
        algo = pg.algorithm(uda)
        algo.set_verbosity(1)
        return algo
    elif name is "ipopt":
        if original_screen_output:
            pl = 5
        else:
            pl = 0
        uda = pg.ipopt()
        uda.set_integer_option("print_level", pl)
        uda.set_integer_option("acceptable_iter", 4)
        uda.set_integer_option("max_iter", 150)

        uda.set_numeric_option("tol", 1e-8)
        uda.set_numeric_option("dual_inf_tol", 1e-8)
        uda.set_numeric_option("constr_viol_tol", 1e-8)
        uda.set_numeric_option("compl_inf_tol", 1e-8)

        uda.set_numeric_option("acceptable_tol", 1e-3)
        uda.set_numeric_option("acceptable_dual_inf_tol", 1e-2)
        uda.set_numeric_option("acceptable_constr_viol_tol", 1e-6)
        uda.set_numeric_option("acceptable_compl_inf_tol", 1e-6)

        algo = pg.algorithm(uda)
        return algo
    elif name is "snopt7":
        import pygmo_plugins_nonfree as pg7
        uda = pg7.snopt7(original_screen_output,
                         "/usr/local/lib/libsnopt7_c.so")
        uda.set_integer_option("Major iterations limit", 2000)
        uda.set_integer_option("Iterations limit", 200000)
        uda.set_numeric_option("Major optimality tolerance", 1e-2)
        uda.set_numeric_option("Major feasibility tolerance", 1e-9)

        algo = pg.algorithm(uda)
        return algo



import numpy as np


def load_gravity_model(fp):
    """Load spherical harmonics model.

    Args:
        - fp (``str``): path to model file.

    Returns:
        - planet_radius (``float``): Radius of central body (m)
        - gravity_param (``float``): Gravitational parameter of central body (m3/s2)
        - c_coefficients (``array-like``): Array of C_(n, m) coefficients.
        - s_coefficients (``array-like``): Array of S_(n, m) coefficients.
        - degree (``int``): Maximum degree of model.
        - order (``int``): Maximum order of model.

    Example::

        r, mu, c, s, n, m = pykep.util.load_gravity_model('gravity_models/Earth/egm96.txt')

    File format:
        The input file must follow a specific format. The first row contains planetary radius in km and
        gravitational parameter in km3/s2. The second row until the end contains normalised C and S 
        coefficients for every degree and order. The file is comma-delimited. Gravity models can be downloaded
        from http://pds-geosciences.wustl.edu/dataserv/gravity_models.htm (NASA)
        and https://earth.esa.int/web/guest/data_access/browse-data-products (ESA).

    
    File::

          1| radius,mu,degree,order
          2| 0,0,C(0,0),S(0,0)
          3| 1,0,C(1,0),S(1,0)
          4| ...        
    """
    # todo: check file

    planet_radius, gravity_param, degree, order, array = _read_model_file(fp)

    c_coefficients, s_coefficients = _convert_data(array, degree, order)

    return planet_radius, gravity_param, c_coefficients, s_coefficients, degree, order


def _read_model_file(fp):
    """Read model data from file"""
    with open(fp) as f:
        header = np.genfromtxt(f, delimiter=",", max_rows=1)

        if len(header) != 4:
            raise IndexError("file header needs at least 4 values: radius, mu, degree, order. " +
                             f"Header contains ({len(header)}) values.")

        radius = float(header[0]) * 10**3
        gravity_const = float(header[1]) * 10**9
        degree = int(header[2])
        order = int(header[3])

        array = np.genfromtxt(f, delimiter=",", skip_header=1)

        return radius, gravity_const, degree, order, array


def _convert_data(array, max_degree, max_order):
    """Convert array from file to two arrays containing only C or S coefficients."""
    c_array = np.zeros((max_degree + 1, max_order + 1))
    s_array = np.zeros((max_degree + 1, max_order + 1))

    for row in array:
        degree = int(row[0])
        order = int(row[1])

        c = row[2]
        s = row[3]

        c_array[degree, order] = c
        s_array[degree, order] = s

    return c_array, s_array



from pykep.trajopt import mga_1dsm, launchers
from pykep.planet import jpl_lp

import numpy as np
from numpy.linalg import norm
from math import log, acos, cos, sin, asin, exp
from copy import deepcopy


class _juice_udp(mga_1dsm):
    """
    This class represents a rendezvous mission to Jupiter modelled as an MGA-1DSM transfer. The selected fly-by sequence,
    E-EVEME-J, and other parameters are inspired to the ESA Juice mission. A luncher model is included, namely an Ariane5
    launch from Kourou.
    JUICE - JUpiter ICy moons Explorer - is the first large-class mission in ESA's Cosmic Vision 2015-2025 programme.
    Planned for launch in 2022 and arrival at Jupiter in 2029, it will spend at least three years making detailed
    observations of the giant gaseous planet Jupiter and three of its largest moons, Ganymede, Callisto and Europa.
    """

    def __init__(self, multi_objective, tof_encoding, tof):
        """
        Args:
            - multi_objective (``bool``): when True the problem fitness will return also the time of flight as an added objective
            - tof_encoding (``str``): one of 'direct', 'eta' or 'alpha'. Selects the encoding for the time of flights
            - tof (``list`` or ``list`` of ``list``): time of flight bounds. As documented in ``pykep.mga_1dsm``

        """
        # Redefining the planets as to change their safe radius
        earth = jpl_lp('earth')
        earth.safe_radius = 1.05
        # We need the Earth eph in the fitnes
        venus = jpl_lp('venus')
        venus.safe_radius = 1.05
        mars = jpl_lp('mars')
        mars.safe_radius = 1.05
        jupiter = jpl_lp('jupiter')

        super().__init__(
            seq=[earth, earth, venus, earth, mars, earth, jupiter],
            t0=[8000, 8400],
            tof=tof,
            vinf=[1., 4.],
            add_vinf_dep=False,
            add_vinf_arr=True,
            tof_encoding=tof_encoding,
            multi_objective=multi_objective,
            orbit_insertion=True,
            e_target=0.98531407996358,
            rp_target=1070400000,
            eta_lb=0.01,
            eta_ub=0.99,
            rp_ub=10)

    def fitness(self, x):
        T, Vinfx, Vinfy, Vinfz = self._decode_times_and_vinf(x)
        # We transform it (only the needed component) to an equatorial system rotating along x
        # (this is an approximation, assuming vernal equinox is roughly x and the ecliptic plane is roughly xy)
        earth_axis_inclination = 0.409072975
        # This is different from the GTOP tanmEM problem, I think it was bugged there as the rotation was in the wrong direction.
        Vinfz = - Vinfy * sin(earth_axis_inclination) + Vinfz * cos(earth_axis_inclination)
        # And we find the vinf declination (in degrees)
        sindelta = Vinfz / x[3]
        declination = asin(sindelta) / np.pi * 180.
        # We now have the initial mass of the spacecraft
        m_initial = launchers.ariane5(x[3] / 1000., declination)
        # And we can evaluate the final mass via Tsiolkowsky
        Isp = 312.
        g0 = 9.80665

        if self._multi_objective:
            DV, T = super().fitness(x)
        else:
            DV, = super().fitness(x)

        DV = DV + 275.  # losses for 5 swingbys + insertion
        m_final = m_initial * exp(-DV / Isp / g0)
        # Numerical guard for the exponential
        if m_final == 0:
            m_final = 1e-320

        encoded_m_final = -log(m_final)

        if self._multi_objective:
            return (encoded_m_final, T)

        return (encoded_m_final,)

    def get_name(self):
        return "Juice (Trajectory Optimization Gym P13-14)"

    def __repr__(self):
        return self.get_name()

    def get_extra_info(self):
        retval = "\t Sequence: " + \
            [pl.name for pl in self._seq].__repr__()
        return retval

    def pretty(self, x):
        """
        prob.plot(x)

        - x: encoded trajectory

        Prints human readable information on the trajectory represented by the decision vector x

        Example::

          print(prob.pretty(x))
        """
        super().pretty(x)
        T, Vinfx, Vinfy, Vinfz = self._decode_times_and_vinf(x)
        # We transform it (only the needed component) to an equatorial system rotating along x
        # (this is an approximation, assuming vernal equinox is roughly x and the ecliptic plane is roughly xy)
        earth_axis_inclination = 0.409072975
        # This is different from the GTOP tanmEM problem, I think it was bugged there as the rotation was in the wrong direction.
        Vinfz = - Vinfy * sin(earth_axis_inclination) + Vinfz * cos(earth_axis_inclination)
        # And we find the vinf declination (in degrees)
        sindelta = Vinfz / x[3]
        declination = asin(sindelta) / np.pi * 180.
        m_initial = launchers.ariane5(x[3] / 1000., declination)
        # And we can evaluate the final mass via Tsiolkowsky
        Isp = 312.
        g0 = 9.80665
        DV = super().fitness(x)[0]
        DV = DV + 275.  # losses for 5 swgbys + insertion
        m_final = m_initial * exp(-DV / Isp / g0)
        print("\nInitial mass:", m_initial)
        print("Final mass:", m_final)
        print("Declination:", declination)


# Problem P13: JUICE mission MGA1DSM, single objective, direct encoding
juice = _juice_udp(
    multi_objective=False,
    tof_encoding='direct',
    tof=[[200, 500], [30, 300], [200, 500], [30, 300], [500, 800], [900, 1200]])

# Problem P14: JUICE mission MGA1DSM, multiple objective, alpha encoding
juice_mo = _juice_udp(
    multi_objective=True,
    tof_encoding='alpha',
    tof=[2000, 3000])



from pykep.trajopt import pl2pl_N_impulses
from pykep.planet import jpl_lp


class _emNimp_udp(pl2pl_N_impulses):
    """
    Write Me
    """
    def __init__(self, N=3):
        """
        Write Me
        """
        super().__init__(
            start=jpl_lp('earth'),
            target=jpl_lp('mars'),
            N_max=N,
            tof=[200., 700.],
            vinf=[0., 4.],
            phase_free=False,
            multi_objective=False,
            t0=[10000, 11000]
        )
        self.N = N

    def get_name(self):
        return "Earth-Mars " + str(self.N) + " impulses (Trajectory Optimisation Gym P" + str(int(4 + (self.N - 3.)/2.)) + ")"

    def get_extra_info(self):
        retval = "\tTrajectory Optimisation Gym problem (P"+ str(int(4 + (self.N - 3.)/2.)) +"): Earth-Mars, " + str(
            int(4 + (self.N - 3.)/2.)) + " impulses, single objective\n"
        return retval

    def __repr__(self):
        return self.get_name()


# Problem P4: Earth-Mars, 3 impulses, single objective, alpha encoding
em3imp = _emNimp_udp(N=3)
# Problem P5: Earth-Mars, 5 impulses, single objective, alpha encoding
em5imp = _emNimp_udp(N=5)
# Problem P6: Earth-Mars, 7 impulses, single objective, alpha encoding
em7imp = _emNimp_udp(N=7)



class knn():
    """
    The class finds the k-nearest neighbours to a given planet from a list of planets.
    The problem of finding who is "close-by" can be efficiently solved using appropriate data structures.
    Here a kdtree is employed bringng complexity down to O(log N). The k-d-tree can then be queried efficiently
    for all asteroid within a given distance ('ball' query) or for all k closest asteroids ('knn' query).

    The notion of distance used (metric) can be:

    - 'euclidean': the simple Euclidean distance over the asteroid's (r,v). 
      The position an velocity vectors are scaled w.r.t. some reference values.

    - 'orbital', the distance is computed with respect to (r/T + v, r/T), corresponding to the DV computed over
      a linear model of the orbital transfer. The distance returned will thus be in m/s.

    """

    from pykep.core import AU, EARTH_VELOCITY

    def _eph_normalize(self, eph):
        """
        Normalize a body's ephemerides:
            ( r / refn_r, v / refn_v )
        where `refn_r` and `refn_v` are reference values you can set in the
        arguments (defaults: refn_r == pk.AU, refn_v == pk.EARTH_VELOCITY).

        Accepts either a single ephemeride, or a matrix with one per row.

        Example:
        >>> eph_normalize( asteroid_list[0].eph(TRAJ_START_MIN) )
        array([ -1.80619704e-01,   9.66554711e-01,  -1.46653866e-05,
                -9.99416814e-01,  -1.87484002e-01,   4.12493855e-06])
        """
        import numpy as np

        if type(eph) == tuple:
            eph = np.hstack(eph)
        e = eph.reshape(-1, 6)

        # normalize r
        e[:, :3] /= self._ref_r
        # normalize v
        e[:, -3:] /= self._ref_v

        return e.reshape(eph.shape)

    def _orbital_metric(self, r, v):
        from pykep.core import DAY2SEC
        DV2 = [a / (self._T * DAY2SEC) for a in r]
        DV1 = [a + b for a, b in zip(DV2, v)]
        return (DV1, DV2)

    def _make_kdtree(self, t):
        """
        Returns a kd-tree data structure indexing the normalized ephemerides
        of planets_list, at epoch t.

        make_kdtree( planets_list, t, ref_r=AU, ref_v=EARTH_VELOCITY )

        - planets_list: list of pykep.planet objects
        - t: epoch

        The returned kd-tree can then be used for efficient nearest-neighbor queries.

        See also:
            http://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.html
            https://en.wikipedia.org/wiki/K-d_tree

        Examples::

          ast_list = [pykep.planet_gtoc7(idx) for idx in range(1,16256)]
          kdt = pykep.phasing.make_kdtree(ast_list,epykep.epoch(6754.3))
        """

        from scipy.spatial import cKDTree
        import numpy as np

        if self._metric == 'euclidean':
            e = np.array([a.eph(t) for a in self._asteroids])
        else:
            e = np.array([self._orbital_metric(*a.eph(t))
                          for a in self._asteroids])

        # reshape memory area, so each asteroid's ephemeride gets represented by
        # a single 6 dimensional vector
        e = e.reshape((e.shape[0], -1))

        # normalize the full matrix
        # (if the `normalize_f` is set to None, no normalization takes place)
        if self._metric == 'euclidean':
            self._eph_normalize(e)

        return cKDTree(e)

    def __init__(self, planet_list, t, metric='orbital', ref_r=AU, ref_v=EARTH_VELOCITY, T=180.0):
        """
        USAGE: knn = knn(planet_list, t, metric='orbital', ref_r=AU, ref_v=EARTH_VELOCITY, T=365.25):

        - planet_list   list of pykep planets (typically thousands)
        - t             epoch
        - metric        one of ['euclidean', 'orbital']
        - ref_r         reference radius   (used as a scaling factor for r if the metric is 'euclidean')
        - ref_v         reference velocity (used as a scaling factor for v if the metric is 'euclidean')
        - T             average transfer time (used in the definition of the 'orbital' metric)

        Example::

            from pykep import *
            pl_list = [planet.gtoc7(i) for i in range(16257)]
            knn = phasing.knn(pl_list, epoch(t0), metric='orbital', T=180)
            neighb, ids, dists = knn.find_neighbours(pl_list[ast_0], query_type='knn', k=10000)
            neighb, ids, _ = knn.find_neighbours(pl_list[ast_0], query_type='ball', r=5000)
        """
        import numpy as np
        import pykep as pk
        self._asteroids = np.array(planet_list, dtype=np.object)
        self._ref_r = ref_r
        self._ref_v = ref_v
        self._t = t
        self._metric = metric
        self._T = T
        self._kdtree = self._make_kdtree(self._t)

    def find_neighbours(self, query_planet, query_type='knn', *args, **kwargs):
        """
        Finds the neighbours of a given planet at a given epoch. The user may query for the
        k-nearest neighbours or for all neighbours within a given distance

        knn.find_neighbours(query_planet, query_type='knn', \*args, \*\*kwargs )

        - query_planet: the planet we want to find neighbours of. Can be an integer, in which case it refers to the idx in self.asteroid_list
        - query_type: one of 'knn' or 'ball'.
        - \*args, \*\*args: according to the query type (read below)

        Returns (neighb, neighb_ids, dists), where dist is only computed if 'knn' type query is made

        The following kinds of spatial queries are currently implemented:

        query_type = 'knn':
            The kwarg 'k' determines how many k-nearest neighbours are returned
            For arguments, see:
            http://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.query.html

        query_type = 'ball':
            The kwarg 'r' determines the distance within which all asteroids are returned.
            For arguments, see:
            http://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.query_ball_point.html
        """
        if type(query_planet) == int:
            query_planet = self._asteroids[query_planet]

        # generate the query vector
        x = query_planet.eph(self._t)
        if self._metric == 'euclidean':
            x = self._eph_normalize(x)
        else:
            DV1, DV2 = self._orbital_metric(x[0], x[1])
            x = DV1 + DV2

        if query_type == 'knn':
            # Query for the k nearest neighbors
            # http://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.query.html
            dists, idxs = self._kdtree.query(x, *args, **kwargs)
        elif query_type == 'ball':
            # Query for all neighbors within a sphere of given radius
            # http://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.query_ball_point.html
            idxs = self._kdtree.query_ball_point(x, *args, **kwargs)
            dists = [None] * len(idxs)
        else:
            raise Exception('Unrecognized query type: %s' % str(query_type))

        neighb = [
            # (ast. object, ast. ID, distance)
            (self._asteroids[i], i, d)
            for i, d in zip(idxs, dists)
        ]

        # split into three lists, one of objects, one of IDs, and one for
        # distances
        neighb, neighb_ids, dists = list(
            zip(*neighb)) if neighb != [] else ([], [], [])

        return neighb, neighb_ids, dists



from pykep.trajopt import mga_1dsm
from pykep.planet import jpl_lp, keplerian
from pykep import AU, DEG2RAD, MU_SUN, epoch

# MESSENGER (FULL) (we need to modify the safe radius of the planets to match the wanted problem)
_earth = jpl_lp('earth')
_venus = jpl_lp('venus')
_venus.safe_radius = 1.1
_mercury = jpl_lp('mercury')
_mercury.safe_radius = 1.05
_seq_messenger = [_earth,
                _venus,
                _venus,
                _mercury,
                _mercury,
                _mercury,
                _mercury
]

class _messenger_udp(mga_1dsm):
    """
    This class represents a rendezvous mission to Mercury modelled as an MGA-1DSM transfer. The selected fly-by sequence,
    E-VVMeMeMe-Me, and other parameters are inspired to the Messenger mission. We have only omitted the first Earth fly-by that
    was used to correct for launcher performances, since we here do not make use of a luncher model.
    As far as chemical propelled interplanetary trajectories go, this particular one is particularly complex and difficult
    to design. The time of flights among successive Mercury fly-bys allow for multiple rvolutions and resonances, making
    optimization techniques struggle to find the correct combination.
    The amount of specialistic knowledge that needs to be used to obtain a successfull design is significant.
    Finding a global optimization approach able to find a good trajectory in complete autonomy without making
    use of additional problem knowledge is possible, but limiting the number of fitness call is difficult.

    .. note::

       A significantly similar version of this problem was part of the no longer maintained GTOP database, 
       https://www.esa.int/gsp/ACT/projects/gtop/gtop.html. The exact definition is, though, different and results
       cannot thus not be compared to those posted in GTOP.
    """
    def __init__(self):
        """
        Write Me
        """
        super().__init__(
            seq =_seq_messenger,
            t0 = [1900, 2300],
            tof = [[100, 500], [100, 500], [100, 500], [100, 500], [100, 500], [100, 600]],
            vinf = [2.5, 4.05],
            add_vinf_dep = False,
            add_vinf_arr = True,
            tof_encoding = "direct",
            multi_objective = False,
            orbit_insertion = True,
            e_target = 0.704,
            rp_target = 2640000,
            eta_lb = 0.01,
            eta_ub = 0.99,
            rp_ub = 6.
        )

# Problem P15: Messenger mission MGA1DSM, single objective, direct encoding
messenger = _messenger_udp()




def run_example5():
    import pygmo as pg
    from pykep import epoch
    from pykep.planet import jpl_lp
    from pykep.trajopt import mga_1dsm

    # We define an Earth-Venus-Earth problem (single-objective)
    seq = [jpl_lp('earth'), jpl_lp('venus'), jpl_lp('earth')]
    udp = mga_1dsm(
        seq=seq,
        t0=[epoch(5844), epoch(6209)],
        tof=[0.7 * 365.25, 3 * 365.25],
        vinf=[0.5, 2.5],
        add_vinf_dep=False,
        add_vinf_arr=True,
        multi_objective=False
    )

    pg.problem(udp)
    # We solve it!!
    uda = pg.sade(gen=100)
    archi = pg.archipelago(algo=uda, prob=udp, n=8, pop_size=20)
    print(
        "Running a Self-Adaptive Differential Evolution Algorithm .... on 8 parallel islands")
    archi.evolve(10)
    archi.wait()
    sols = archi.get_champions_f()
    idx = sols.index(min(sols))
    print("Done!! Solutions found are: ", archi.get_champions_f())
    udp.pretty(archi.get_champions_x()[idx])
    udp.plot(archi.get_champions_x()[idx])

if __name__ == "__main__":
    run_example5()



def run_example2():
    import matplotlib as mpl
    from mpl_toolkits.mplot3d import Axes3D

    import matplotlib.pyplot as plt
    from pykep import epoch, DAY2SEC, AU, MU_SUN, lambert_problem
    from pykep.planet import jpl_lp
    from pykep.orbit_plots import plot_planet, plot_lambert

    mpl.rcParams['legend.fontsize'] = 10

    fig = plt.figure()
    axis = fig.gca(projection='3d')

    t1 = epoch(0)
    t2 = epoch(640)
    dt = (t2.mjd2000 - t1.mjd2000) * DAY2SEC

    axis.scatter([0], [0], [0], color='y')

    pl = jpl_lp('earth')
    plot_planet(
        pl, t0=t1, color=(0.8, 0.8, 1), legend=True, units=AU, axes=axis)
    rE, _ = pl.eph(t1)

    pl = jpl_lp('mars')
    plot_planet(
        pl, t0=t2, color=(0.8, 0.8, 1), legend=True, units=AU, axes=axis)
    rM, _ = pl.eph(t2)

    l = lambert_problem(rE, rM, dt, MU_SUN)
    plot_lambert(l, color='b', legend=True, units=AU, axes=axis)
    plot_lambert(l, sol=1, color='g', legend=True, units=AU, axes=axis)
    plot_lambert(l, sol=2, color='g', legend=True, units=AU, axes=axis)

    plt.show()



import pykep as pk


class mr_lt_nep:
    """
    This class represents, as a global optimization problem (linearly constrained,
    high diemensional), a Multiple Randezvous trajectory of a low-thrust spacecraft equipped
    with a nuclear electric propulsion engine.

    - Izzo, D. et al., GTOC7 - Solution Returned by the ACT/ISAS team

    The decision vector (chromosome) is:
    [t1, tof1, rest1, m_f1] + [throttles1] +
    [t2, tof2, rest2, m_f2] + [throttles2] + ....
    .... + [total_tof]

    where the units are [mjd2000, days, days, kg] + [n/a] + .... + [days]

    .. note::

       The resulting problem is non linearly constrained. The resulting trajectory is not time-bounded.
    """
    def __init__(
            self,
            seq=[pk.planet.gtoc7(3413), pk.planet.gtoc7(
                234), pk.planet.gtoc7(11432)],
            n_seg=5,
            t0=[13000, 13200],
            leg_tof=[1, 365.25 * 3],
            rest=[30., 365.25],
            mass=[800, 2000],
            Tmax=0.3,
            Isp=3000.,
            traj_tof=365.25 * 6,
            objective='mass',
            c_tol=1e-05
    ):
        """
        prob = mr_lt_nep(seq=[pykep.gtoc7(3413),pykep.gtoc7(234), pykep.gtoc7(11432)], n_seg=5, t0=[13000, 13200],
                    leg_tof=[1, 365.25 * 3], rest=[30., 365.25], mass=[800, 2000], Tmax=0.3,
                    Isp=3000., traj_tof=365.25 * 6, objective='mass', c_tol=1e-05)

        * seq: list of pykep.planet defining the encounter sequence for the trajectoty (including the initial planet)
        * n_seg: list of integers containing the number of segments to be used for each leg (len(n_seg) = len(seq)-1)
        * t0: list of two pykep epochs defining the launch window
        * leg_tof: list of two floats defining the minimum and maximum time of each leg (days)
        * rest: list of two floats defining the minimum and maximum time the spacecraft can rest at one planet (days)
        * mass: list of two floats defining the minimum final spacecraft mass and the starting spacecraft mass (kg)
        * Tmax: maximum thrust (N)
        * Isp: engine specific impulse (sec)
        * traj_tof maximum total mission duration (days)
        * c_tol: tolerance on the constraints
        """

        # Number of legs
        n = len(seq) - 1
        # Problem dimension
        dim = (4 + n_seg * 3) * n + 1
        # Number of equality constraints
        self.__dim_eq = 7 * n
        # Number of Inequality constraints
        self.__dim_ineq = n * n_seg + n

        # We define data members
        self.__seq = seq
        self.__num_legs = n
        self.__nseg = n_seg
        self.__dim_leg = 4 + n_seg * 3
        self.__start_mass = mass[1]
        self.__max_total_time = traj_tof
        self.__t0 = t0
        self.__leg_tof = leg_tof
        self.__rest = rest
        self.__mass = mass

        # We create n distinct legs objects
        self.__legs = []
        for i in range(n):
            self.__legs.append(pk.sims_flanagan.leg())
        for leg in self.__legs:
            leg.high_fidelity = True
            leg.set_mu(pk.MU_SUN)

        if objective not in ['mass', 'time']:
            raise ValueError(
                "Error in defining the objective. Was it one of mass or time?")
        self.__objective = objective

    def fitness(self, x_full):
        retval = []
        # 1 - obj fun
        if self.__objective == 'mass':
            retval.append(x_full[-1 - self.__dim_leg + 3])
        elif self.__objective == 'time':
            retval.append(x_full[-1 - self.__dim_leg] - x_full[0])

        sc_mass = self.__start_mass
        eqs = []
        ineqs = []

        # 2 - constraints
        for i in range(self.__num_legs):
            x = x_full[i * self.__dim_leg:(i + 1) * self.__dim_leg]

            start = pk.epoch(x[0])
            end = pk.epoch(x[0] + x[1])

            # Computing starting spaceraft state
            r, v = self.__seq[i].eph(start)
            x0 = pk.sims_flanagan.sc_state(r, v, sc_mass)

            # Computing ending spaceraft state
            r, v = self.__seq[i + 1].eph(end)
            xe = pk.sims_flanagan.sc_state(r, v, x[3])

            # Building the SF leg
            self.__legs[i].set_spacecraft(
                pk.sims_flanagan.spacecraft(sc_mass, .3, 3000.))
            self.__legs[i].set(start, x0, x[-3 * self.__nseg:], end, xe)

            # Setting all constraints
            eqs.extend(self.__legs[i].mismatch_constraints())
            ineqs.extend(self.__legs[i].throttles_constraints())

            eqs[-7] /= pk.AU
            eqs[-6] /= pk.AU
            eqs[-5] /= pk.AU
            eqs[-4] /= pk.EARTH_VELOCITY
            eqs[-3] /= pk.EARTH_VELOCITY
            eqs[-2] /= pk.EARTH_VELOCITY
            eqs[-1] /= self.__start_mass

            sc_mass = x[3]  # update mass to final mass of leg

            if i < self.__num_legs - 1:
                x_next = x_full[
                    (i + 1) * self.__dim_leg:(i + 2) * self.__dim_leg]
                time_ineq = x[0] + x[1] + x[2] - x_next[0]
                ineqs.append(time_ineq / 365.25)
            else:
                final_time_ineq = x[0] + x[1] + x[2] - \
                    x_full[0] - x_full[-1]  # <- total time
                ineqs.append(final_time_ineq / 365.25)

        retval = retval + eqs + ineqs
        return retval

    def get_bounds(self):
        t0 = self.__t0
        leg_tof = self.__leg_tof
        rest = self.__rest
        mass = self.__mass
        nseg = self.__nseg
        traj_tof = self.__max_total_time
        n = self.__num_legs
        # We set the ptoblem box-bounds
        # set leg bounds
        lb_leg = [t0[0], leg_tof[0], rest[0], mass[0]] + [-1] * nseg * 3
        ub_leg = [t0[1] + traj_tof * n, leg_tof[1],
                  rest[1], mass[1]] + [1] * nseg * 3

        # set n leg bounds
        lb = lb_leg * n
        ub = [t0[1], leg_tof[1], rest[1], mass[1]] + \
            [1] * nseg * 3 + ub_leg * (n - 1)

        # set total time bounds
        lb += [1.]
        ub += [self.__max_total_time]

        return (lb, ub)

    def get_nic(self):
        return self.__dim_ineq

    def get_nec(self):
        return self.__dim_eq

    def resting_times(self, x):
        return list(x[2::self.__dim_leg])

    def plot(self, x, ax=None):
        """
        ax = prob.plot(x, ax=None)

        - x: encoded trajectory
        - ax: matplotlib axis where to plot. If None figure and axis will be created
        - [out] ax: matplotlib axis where to plot

        Plots the trajectory represented by a decision vector x on the 3d axis ax

        Example::

          ax = prob.plot(x)
        """
        import matplotlib as mpl
        from mpl_toolkits.mplot3d import Axes3D
        import matplotlib.pyplot as plt
        from pykep import epoch, AU
        from pykep.sims_flanagan import sc_state
        from pykep.orbit_plots import plot_planet, plot_sf_leg

        # Creating the axis if necessary
        if ax is None:
            mpl.rcParams['legend.fontsize'] = 10
            fig = plt.figure()
            axis = fig.gca(projection='3d')
        else:
            axis = ax

        # Plotting the Sun ........
        axis.scatter([0], [0], [0], color='y')

        # Plotting the pykep.planet both at departure and arrival dates
        for i in range(self.__num_legs):
            idx = i * self.__dim_leg
            plot_planet(self.__seq[i], epoch(x[idx]), units=AU, legend=True, color=(
                0.7, 0.7, 0.7), s=30, axes=axis)
            plot_planet(self.__seq[i + 1], epoch(x[idx] + x[idx + 1]),
                        units=AU, legend=False, color=(0.7, 0.7, 0.7), s=30, axes=axis)

        # Computing the legs
        self.fitness(x)

        # Plotting the legs
        for leg in self.__legs:
            plot_sf_leg(leg, units=AU, N=10, axes=axis, legend=False)

        return axis



# -*- coding: utf-8 -*-

# Copyright 2017-2018 PaGMO development team
#
# This file is part of the PaGMO library.
#
# The PaGMO library is free software; you can redistribute it and/or modify
# it under the terms of either:
#
#   * the GNU Lesser General Public License as published by the Free
#     Software Foundation; either version 3 of the License, or (at your
#     option) any later version.
#
# or
#
#   * the GNU General Public License as published by the Free Software
#     Foundation; either version 3 of the License, or (at your option) any
#     later version.
#
# or both in parallel, as here.
#
# The PaGMO library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received copies of the GNU General Public License and the
# GNU Lesser General Public License along with the PaGMO library.  If not,
# see https://www.gnu.org/licenses/.

from __future__ import absolute_import as _ai

import unittest as _ut
import numpy as np
import pathlib

from pykep import __extensions__


class core_functions_test_case(_ut.TestCase):
    """Test case for the core functions

    """

    def runTest(self):
        self.run_epoch_test()

    def run_epoch_test(self):
        from .core import epoch
        epoch(julian_date=0, julian_date_type='mjd2000')


class lambert_test_case(_ut.TestCase):
    """Test case for the lambert problem class

    """

    def runTest(self):
        from .core import lambert_problem
        lambert_problem(r1=[1, 1, 0], r2=[0, 1, 0],
                        mu=1., cw=False, max_revs=0, tof=0.3)


class mga_1dsm_test_case(_ut.TestCase):
    """Test case for the mga1_dsm class

    """

    def runTest(self):
        self.run_construction_test()
        self.run_decode_times_and_vinf_test()

    def run_construction_test(self):
        from .trajopt import mga_1dsm
        # Correct use (nothrow)
        udp = mga_1dsm()
        self.assertEqual(udp.get_nobj(), 1)
        udp = mga_1dsm(tof_encoding='direct', tof=[[20, 400], [20, 400]])
        self.assertEqual(udp.get_nobj(), 1)
        udp = mga_1dsm(tof_encoding='eta', tof=500)
        self.assertEqual(udp.get_nobj(), 1)
        udp = mga_1dsm(tof_encoding='alpha', tof=[20, 500])
        self.assertEqual(udp.get_nobj(), 1)
        udp = mga_1dsm(tof_encoding='direct', tof=[
                       [20, 400], [20, 400]], multi_objective=True)
        self.assertEqual(udp.get_nobj(), 2)

        # Incorrect use (raise)
        with self.assertRaises(TypeError):
            mga_1dsm(tof_encoding='direct', tof=34)
        with self.assertRaises(TypeError):
            mga_1dsm(tof_encoding='direct', tof=[[400], [20, 400]])
        with self.assertRaises(TypeError):
            mga_1dsm(tof_encoding='direct', tof=[20, 400])
        with self.assertRaises(TypeError):
            mga_1dsm(tof_encoding='eta', tof=[20, 400])
        with self.assertRaises(TypeError):
            mga_1dsm(tof_encoding='eta', tof=[[20, 400], [20, 400]])
        with self.assertRaises(TypeError):
            mga_1dsm(tof_encoding='alpha', tof=4)
        with self.assertRaises(TypeError):
            mga_1dsm(tof_encoding='alpha', tof=[[20, 400], [20, 400]])

    def run_decode_times_and_vinf_test(self):
        from .trajopt import mga_1dsm
        # Alpha
        udp = mga_1dsm(tof_encoding='alpha', tof=[20, 500])
        x = [10] + [0., 0., 1., 1., 0.5] + [0.4, 1.3, 0.5, 0.5] + [50]
        retval = udp._decode_times_and_vinf(x)
        self.assertAlmostEqual(retval[0][0], 25.)
        self.assertAlmostEqual(retval[0][1], 25.)
        self.assertAlmostEqual(retval[1], 0.)
        self.assertAlmostEqual(retval[2], 0.)
        self.assertAlmostEqual(retval[3], 1.)
        # Eta
        udp = mga_1dsm(tof_encoding='eta', tof=50)
        x = [10] + [0., 0., 1., 1., 0.5] + [0.4, 1.3, 0.5, 0.5]
        retval = udp._decode_times_and_vinf(x)
        self.assertAlmostEqual(retval[0][0], 25.)
        self.assertAlmostEqual(retval[0][1], 12.5)
        self.assertAlmostEqual(retval[1], 0.)
        self.assertAlmostEqual(retval[2], 0.)
        self.assertAlmostEqual(retval[3], 1.)
        # Direct
        udp = mga_1dsm(tof_encoding='direct', tof=[[10, 400], [10, 400]])
        x = [10] + [0., 0., 1., 1., 123] + [0.4, 1.3, 0.5, 321]
        retval = udp._decode_times_and_vinf(x)
        self.assertAlmostEqual(retval[0][0], 123)
        self.assertAlmostEqual(retval[0][1], 321)
        self.assertAlmostEqual(retval[1], 0.)
        self.assertAlmostEqual(retval[2], 0.)
        self.assertAlmostEqual(retval[3], 1.)


class gym_test_case(_ut.TestCase):
    """Test case for the gym

    """

    def runTest(self):
        self.run_rosetta_test()
        self.run_cassini2_test()
        self.run_tandem_test()
        self.run_juice_test()
        self.run_messenger_test()

    def run_rosetta_test(self):
        from .trajopt import gym
        udp = gym.rosetta
        x = [1.53488329e+03, 4.56388378e-01, 9.51717655e-01, 4.18212047e+03,
             4.32159299e-01, 3.65256539e+02, 5.03363275e+00, 2.38949977e+00,
             4.55746823e-01, 7.09999954e+02, 1.79894273e+00, 1.05000003e+00,
             6.09083347e-01, 2.60816142e+02, 4.95158968e+00, 3.16049580e+00,
             6.89049263e-01, 7.29775762e+02, 4.30823655e+00, 1.10842692e+00,
             4.16075410e-01, 1.84999995e+03]
        self.assertAlmostEqual(udp.fitness(x)[0], 1371.4992595125018)

    def run_cassini2_test(self):
        from .trajopt import gym
        udp = gym.cassini2
        x = [-7.75699976e+02,  9.15777367e-01,  4.06442043e-01,  3.21309562e+03,
        6.81118341e-01,  1.62660490e+02, -1.58051063e+00,  1.28479507e+00,
        4.72699902e-01,  4.24319550e+02,  4.30475919e+00,  1.15739933e+00,
        2.55718252e-01,  5.44489098e+01, -1.54332794e+00,  1.27160729e+00,
        9.00000000e-01,  5.88481599e+02,  4.76774269e+00,  7.00000000e+01,
        1.00000000e-02,  2.20000000e+03]
        self.assertAlmostEqual(udp.fitness(x)[0], 1511.731793212)

    def run_tandem_test(self):
        from .trajopt import gym
        udp = gym.tandem(prob_id = 6, constrained=False)
        x = [ 7.98593791e+03,  9.02569236e-01,  3.59510662e-01,  3.31011777e+03,
        1.00000000e-02,  1.68854894e+02, -1.15321363e+00,  1.05006538e+00,
        4.21911761e-01,  3.38579352e+02, -1.68143357e+00,  1.10121434e+00,
        6.39679220e-01,  1.17925136e+03, -2.02838607e+00,  1.05000000e+00,
        2.31554532e-01,  1.80000707e+03]
        self.assertAlmostEqual(udp.fitness(x)[0], -7.302117213470749)

    def run_juice_test(self):
        from .trajopt import gym
        udp = gym.juice
        x = [ 8.16283083e+03,  6.41922787e-01,  6.51202691e-01,  2.51009414e+03,
        2.97841478e-01,  3.81541370e+02,  9.58572190e-01,  1.53007674e+00,
        3.06125365e-01,  1.49264351e+02,  4.10103546e+00,  2.39297670e+00,
        4.34424957e-01,  3.16066418e+02,  4.33225338e+00,  1.30946367e+00,
        4.52048883e-01,  1.63208108e+02,  5.93850330e-01,  1.34871269e+00,
        2.03288502e-01,  6.52494606e+02, -1.37902374e+00,  1.55482534e+00,
        1.96917559e-01,  1.08471126e+03]
        self.assertAlmostEqual(udp.fitness(x)[0], -7.987614956531155)

    def run_messenger_test(self):
        from .trajopt import gym
        udp = gym.messenger
        x = [ 2.03241398e+03,  6.40762059e-01,  6.63357785e-01,  4.04989271e+03,
        6.63732323e-01,  4.50068524e+02, -3.86553343e+00,  3.52631372e+00,
        5.57888828e-01,  2.24619580e+02, -4.45910441e+00,  1.22736521e+00,
        7.08063036e-01,  2.17965497e+02, -2.47894274e+00,  1.43586128e+00,
        5.88391838e-01,  2.62423586e+02, -2.40594385e-02,  2.45470457e+00,
        7.25370468e-01,  3.58067954e+02,  1.47192632e+00,  1.05000000e+00,
        9.02984391e-01,  5.38436770e+02]
        self.assertAlmostEqual(udp.fitness(x)[0], 5855.8143406005165)


class spherical_harmonics_loader_test_case(_ut.TestCase):
    """Test case for the spherical harmonic gravity file loader function.

    """

    def runTest(self):
        self.run_egm96_test()
        self.run_ggm02c_test()
        self.run_ggm02s_test()
        self.run_jgmro120d_test()
        self.run_glgm3150_test()
        self.run_jgl150q1_test()
        self.run_lpe200_test()
        
    def run_egm96_test(self):
        from .util import load_gravity_model

        req_r = 6378137
        req_mu = 3.986004418e+14

        c20_20 = 4.01448327968e-09
        c100_100 = 1.10930637955e-09

        s100_50 = -1.06362863541e-09
        s250_125 = -8.34356616626e-11

        r, mu, c, s, n, m = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Earth/egm96.txt")

        self.assertEqual(r, req_r)
        self.assertEqual(mu, req_mu)
        
        self.assertEqual(c[20, 20], c20_20)
        self.assertEqual(c[100, 100], c100_100)

        self.assertEqual(s[100, 50], s100_50)
        self.assertEqual(s[250, 125], s250_125)

    def run_ggm02c_test(self):
        from .util import load_gravity_model

        req_r = 6378136.3
        req_mu = 3.986004415e+14

        c20_20 = 3.734698472368e-09
        c100_100 = 8.8043569591782e-10

        s100_50 = -8.2691933615552e-10
        s200_100 = -1.3266576400742e-12

        r, mu, c, s, n, m = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Earth/ggm02c.txt")

        self.assertEqual(r, req_r)
        self.assertEqual(mu, req_mu)
        
        self.assertEqual(c[20, 20], c20_20)
        self.assertEqual(c[100, 100], c100_100)

        self.assertEqual(s[100, 50], s100_50)
        self.assertEqual(s[200, 100], s200_100)

    def run_ggm02s_test(self):
        from .util import load_gravity_model

        req_r = 6378136.3
        req_mu = 3.986004415e+14

        c20_20 = 3.7323168217059e-09
        c100_100 = 1.0356187365035e-09

        s100_50 = -8.3581312522994e-10
        s160_80 = -2.327147027496e-09

        r, mu, c, s, n, m = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Earth/ggm02s.txt")

        self.assertEqual(r, req_r)
        self.assertEqual(mu, req_mu)
        
        self.assertEqual(c[20, 20], c20_20)
        self.assertEqual(c[100, 100], c100_100)

        self.assertEqual(s[100, 50], s100_50)
        self.assertEqual(s[160, 80], s160_80)

    def run_jgmro120d_test(self):
        from .util import load_gravity_model

        req_r = 3396000.0
        req_mu = 4.28283758157561e+13

        c20_20 = -4.767064931643e-07
        c100_100 = -6.238892373683e-09

        s100_50 = -1.191278972881e-08
        s120_60 = 5.598027946471e-10

        r, mu, c, s, n, m = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Mars/jgmro_120d.txt")

        self.assertEqual(r, req_r)
        self.assertEqual(mu, req_mu)
        
        self.assertEqual(c[20, 20], c20_20)
        self.assertEqual(c[100, 100], c100_100)

        self.assertEqual(s[100, 50], s100_50)
        self.assertEqual(s[120, 60], s120_60)

    def run_glgm3150_test(self):
        from .util import load_gravity_model

        req_r = 1738000.0
        req_mu = 4.9002800238e+12

        c20_20 = -1.4913227206107e-08
        c100_100 = -3.0367607155944e-08

        s100_50 = 1.7233711523911e-10
        s120_60 = -5.3052488170131e-09

        r, mu, c, s, n, m = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Moon/glgm3_150.txt")

        self.assertEqual(r, req_r)
        self.assertEqual(mu, req_mu)
        
        self.assertEqual(c[20, 20], c20_20)
        self.assertEqual(c[100, 100], c100_100)

        self.assertEqual(s[100, 50], s100_50)
        self.assertEqual(s[120, 60], s120_60)

    def run_jgl150q1_test(self):
        from .util import load_gravity_model

        req_r = 1738000.0
        req_mu = 4.902801076e+12

        c20_20 = 1.17705963027e-07
        c100_100 = -1.5618538762e-08

        s100_50 = 5.1145625569e-09
        s120_60 = -1.05725130139e-09

        r, mu, c, s, n, m = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Moon/jgl_150q1.txt")

        self.assertEqual(r, req_r)
        self.assertEqual(mu, req_mu)
        
        self.assertEqual(c[20, 20], c20_20)
        self.assertEqual(c[100, 100], c100_100)

        self.assertEqual(s[100, 50], s100_50)
        self.assertEqual(s[120, 60], s120_60)

    def run_lpe200_test(self):
        from .util import load_gravity_model

        req_r = 1738000.0
        req_mu = 4.902800238e+12

        c20_20 = -1.5489032082686e-08
        c100_100 = -3.0841142975112e-08

        s100_50 = -1.53194061215089e-09
        s200_100 = 4.25825545837e-09

        r, mu, c, s, n, m = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Moon/lpe_200.txt")

        self.assertEqual(r, req_r)
        self.assertEqual(mu, req_mu)
        
        self.assertEqual(c[20, 20], c20_20)
        self.assertEqual(c[100, 100], c100_100)

        self.assertEqual(s[100, 50], s100_50)
        self.assertEqual(s[200, 100], s200_100)

class gravity_spherical_harmonic_test_case(_ut.TestCase):
    """Test case for the spherical harmonic gravity calculator function.

    """

    def runTest(self):
        from .util import load_gravity_model, gravity_spherical_harmonic

        r, mu, c, s, n_max, m_max = load_gravity_model(pathlib.Path(__file__).parent / "util/gravity_models/Earth/egm96.txt")

        x = np.array([[6.07303362e+06, -1.63535914e-9, -3.22908926e+06],
                      [-5874145.34596, 1745831.60905, 3123338.4834],
                      [5290507.45841, -3377313.30177, -2813012.71391],
                      [-4360347.55688, 4787584.94679, 2318437.92131],
                      [3144590.02052, -5884275.41062, -1672008.17262]])

        acc = gravity_spherical_harmonic(x, r, mu, c, s, n_max, m_max)

        req_acc = np.array([[-7.438268885207450, 4.174587578722027e-05, 3.966055730251360],
                            [7.195259383674340, -2.138389507954604, -3.836583575161607],
                            [-6.482132055071062, 4.138055191303929, 3.456233485081509],
                            [5.344543324379746, -5.868302782861193, -2.849822685569829],
                            [-3.855999728216701, 7.215225587047862, 2.055872629061049]])

        for a, req_a in zip(acc, req_acc):
            for ax, req_ax in zip(a, req_a):
                self.assertAlmostEqual(ax, req_ax, 13)


def run_test_suite(level=0):
    """Run the full test suite.

    This function will raise an exception if at least one test fails.

    Args:
        level(``int``): the test level (higher values run longer tests)

    """

    retval = 0
    suite = _ut.TestLoader().loadTestsFromTestCase(core_functions_test_case)
    suite.addTest(lambert_test_case())
    suite.addTest(mga_1dsm_test_case())
    suite.addTest(gym_test_case())

    if __extensions__['numba']:
        suite.addTest(spherical_harmonics_loader_test_case())
        suite.addTest(gravity_spherical_harmonic_test_case())


    test_result = _ut.TextTestRunner(verbosity=2).run(suite)

    if len(test_result.failures) > 0 or len(test_result.errors) > 0:
        retval = 1
    if retval != 0:
        raise RuntimeError('One or more tests failed.')



def run_example8(nseg=40):
    """
    This example demonstrates the direct method (sims-flanagan) on a planet to planet scenario.
    """
    import pykep as pk
    import pygmo as pg
    import numpy as np
    from matplotlib import pyplot as plt
    from pykep.examples import add_gradient, algo_factory

    # 1 - Algorithm
    algo = algo_factory("snopt7")

    # 2 - Problem
    udp = add_gradient(pk.trajopt.direct_pl2pl(
        p0="earth",
        pf="mars",
        mass=1000,
        thrust=0.1,
        isp=3000,
        vinf_arr=1e-6,
        vinf_dep=3.5,
        hf=False,
        nseg=nseg,
        t0=[1100, 1400],
        tof=[200, 750]),
        with_grad=False
    )

    prob = pg.problem(udp)
    prob.c_tol = [1e-5] * prob.get_nc()

    # 3 - Population
    pop = pg.population(prob, 1)

    # 4 - Solve the problem (evolve)
    pop = algo.evolve(pop)

    # 5 - Inspect the solution
    if prob.feasibility_x(pop.champion_x):
        print("Optimal Found!!")
    else:
        print("No solution found, try again :)")

    udp.udp_inner.pretty(pop.champion_x)

    axis = udp.udp_inner.plot_traj(pop.champion_x)
    plt.title("The trajectory in the heliocentric frame")
    axis = udp.udp_inner.plot_control(pop.champion_x)
    plt.title("The control profile (throttle)")

    plt.ion()
    plt.show()


if __name__ == "__main__":
    run_example8()



from pykep.trajopt import mga_1dsm
from pykep.planet import jpl_lp, keplerian
from pykep import AU, DEG2RAD, MU_SUN, epoch

# CASSINI2 (we need to modify the safe radius of the planets to match the wanted problem)
_earth_cassini2 = jpl_lp('earth')
_earth_cassini2.safe_radius = 1.15
_venus_cassini2 = jpl_lp('venus')
_venus_cassini2.safe_radius = 1.05
_jupiter_cassini2 = jpl_lp('jupiter')
_jupiter_cassini2.safe_radius = 1.7
_seq_cassini2 = [_earth_cassini2,
                _venus_cassini2,
                _venus_cassini2,
                _earth_cassini2,
                _jupiter_cassini2,
                jpl_lp('saturn')]


class _cassini2_udp(mga_1dsm):
    """
    Write Me
    """
    def __init__(self):
        """
        Write Me
        """
        super().__init__(
            seq =_seq_cassini2,
            t0 = [-1000, 0],
            tof = [[100, 400], [100, 500], [30, 300], [400, 1600], [800, 2200]],
            vinf = [3., 5.],
            add_vinf_dep = False,
            add_vinf_arr = True,
            tof_encoding = "direct",
            multi_objective = False,
            orbit_insertion=True,
            e_target=0.98,
            rp_target=108950000,
            eta_lb = 0.01,
            eta_ub = 0.9,
            rp_ub = 70.
        )

    def get_name(self):
        return "Cassini 2 (Trajectory Optimisation Gym P11)"

    def __repr__(self):
        return self.get_name()

# Problem P11: Cassini mission MGA1DSM, single objective, direct encoding
cassini2 = _cassini2_udp()